import os
from ac_segmentation.postprocess.skeletonize_array_uint8_prob import run


def test_seg_skel_uint8(npy_file):
    run(probability_input_path=npy_file+"test_zarr.npy.gz", 
        skeleton_output_path=npy_file+"output_skel.swcs.tar.gz",probability_threshold=0.05,label_size_threshold=80)
            
            
            


import os
import numpy as np
import pandas as pd
import argschema as ags
import json
import pathlib
import posixpath
from scipy.ndimage import affine_transform
import tifffile as tif

class PreprocessParameters(ags.ArgSchema):
    input_dir = ags.fields.InputDir(required=True, description='Input directory')
    output_dir = ags.fields.OutputDir(required=True, description='Output directory')
    chunk_size = ags.fields.Int(dtype='int', required=False, default=416)
    overlap = ags.fields.Int(dtype='int', required=False, default=16)

def chunk_volume(indir, outdir, chunk_size, overlap, mip=0):
    dir_list = ['chunks', 'deskewed', 'inputs']
    for d in dir_list:
        savedir = os.path.join(outdir, d)
        if not os.path.isdir(savedir):
            os.mkdir(savedir)    
    fullres_ds = miplevel_from_group_dir(indir, mip=mip)
    
    # Get volume size from metadata
    mfile = os.path.join(indir, 's%d'%mip, 'attributes.json')
    with open(mfile,'r') as f:
        data = json.load(f)
    volume_size = data['dimensions']
    volume_size.reverse() # xyz -> zyx
    idx1 = np.arange(0, volume_size[0], chunk_size - overlap)
    idx2 = idx1 + chunk_size
    num_chunks = np.min(np.where(volume_size[0] <= idx2)[0]) + 1
    idx1 = idx1[:num_chunks]
    idx2 = idx2[:num_chunks]
    
    for i in range(num_chunks):
        stack = fullres_ds[idx1[i]:idx2[i], ...]
        if idx2[i] > volume_size[0]:
            stack = np.pad(stack, ((0, idx2[i] - volume_size[0]),(0, 0), (0, 0)), 'edge')
        mip_xy = np.max(stack,0) # save mip
        tif.imwrite(os.path.join(outdir, dir_list[0], 'MAX_xy_chunk%02d.tif'%i), mip_xy)     
#         chunk_dir = os.path.join(outdir, dir_list[0], 'chunk%02d'%i)
#         if not os.path.isdir(chunk_dir):
#             os.mkdir(chunk_dir)
#         for j in range(stack.shape[0]):
#             tif.imwrite(os.path.join(chunk_dir, '%03d.tif'%j), stack[j,:,:])
        
        # Deskew stack
        chunk_dir = os.path.join(outdir, dir_list[1], 'chunk%02d'%i)
        if not os.path.isdir(chunk_dir):
            os.mkdir(chunk_dir)        
        stack = deskew_stack(stack, chunk_dir)
        mip_xy = np.max(stack,0) # save mip
        tif.imwrite(os.path.join(outdir, dir_list[1], 'MAX_xy_chunk%02d.tif'%i), mip_xy) 
        
        # Create input 
        file_name = os.path.join(outdir, dir_list[2], 'chunk%02d.tif'%i)
        stack = create_input(stack, file_name, Imax=32640)
        mip_xy = np.max(stack,0) # save mip
        tif.imwrite(os.path.join(outdir, dir_list[2], 'MAX_xy_chunk%02d.tif'%i), mip_xy)

        # Create bounding box for segmentation
        if i == 0:
            stack_size = stack.shape
            bounding_box = [0,0,0, stack_size[2], stack_size[1], stack_size[0]]
            file_name = os.path.join(outdir, dir_list[2], 'bbox.csv')
            pd.DataFrame({'bounding_box':bounding_box}).to_csv(file_name)

def deskew_stack(stack_in, chunkdir=None, pxl_xy=0.406, angle=32.1, step_stage=2, shift=-1):
    # shift=1 (right to left), shift=-1 (left to right)
    h_xz = int(np.round(shift*np.cos(angle*np.pi/180.0)*step_stage/pxl_xy))
    stack_in_size = stack_in.shape

    # Shear transform
    shear_affine = np.eye(4)
    shear_affine[2,0] = h_xz
    if h_xz < 0:
        dx = int(np.ceil(-h_xz*(stack_in_size[0]-1))) # add translation for negative h_xz
        shear_affine[2,3] = dx
   
    # Calculate output stack shape
    nz, ny, nx = stack_in_size
    bb_in = [(0, 0, 0, 1), (nz-1, 0, 0, 1), (0, ny-1, 0, 1), (0, 0, nx-1, 1),
            (nz-1, ny-1, 0, 1), (nz-1, 0, nx-1, 1), (0, ny-1, nx-1, 1), 
             (nz-1, ny-1, nx-1, 1)]
    bb_out = list(map(lambda c: shear_affine @ np.array(c), bb_in))
    bb_out = np.concatenate(bb_out).reshape((-1, 4))
    output_shape = np.max(bb_out, axis=0) - np.min(bb_out, axis=0) + 1
    base = 4
    output_shape = (base * np.ceil(output_shape/base)).astype(np.int64)[:3]
    
    # Apply shear transform (order=1-linear, 3-cubic)
    stack_out = affine_transform(stack_in, np.linalg.inv(shear_affine), 
                                 output_shape=output_shape,order=1) 
    stack_out_size = stack_out.shape
    
    if stack_out_size[0] > stack_in_size[0]:
        # remove last (extra) z plane
        stack_out = stack_out[:-1,:,:]
        stack_out_size = stack_out.shape
    
    # Save new stack as multiple tif files
    if chunkdir:
        for j in range(stack_out_size[0]):
            tif.imwrite(os.path.join(chunkdir, '%03d.tif'%j), stack_out[j,:,:])        
    return(stack_out)

def create_input(stack_in, fname, Imax=65535):
   # Convert to 8bit, pad to make xyz size a multiple of 64,64,16 for segmentation
    cx, cy, cz = 64, 64, 16
    stack_in_size = stack_in.shape

    # Convert to 8bit
    coef = int(Imax/255) # linear conversion
    stack_in[stack_in>Imax]=Imax
    stack_in = np.uint8(np.round(stack_in/coef))

    #Pad to make xyz size a multiple of 64,64,16
    stack_out_size = (int(cz*np.ceil(stack_in_size[0]/cz)), 
                      int(cy*np.ceil(stack_in_size[1]/cy)), 
                      int(cx*np.ceil(stack_in_size[2]/cx)))

    stack_out = np.pad(stack_in, ((0, stack_out_size[0] - stack_in_size[0]),
                               (0, stack_out_size[1] - stack_in_size[1]), 
                               (0, stack_out_size[2] - stack_in_size[2])), 'edge')

    # Save stack as single tif file
    tif.imwrite(fname, stack_out) 
    return stack_out   
    

# General purpose helper methods
def ldel(string_to_edit, del_string):
    """delete a leading string from another string"""
    slcstart = (
        len(del_string) if string_to_edit.startswith(del_string)
        else 0)
    return string_to_edit[slcstart:]


def iter_parents(path):
    """iterate over the nearest parent paths of a given path"""
    path = pathlib.Path(path)
    yield path.parent
    yield from iter_parents(path.parent)


# General n5 or bdv.n5 utility methods
def get_n5_dir(group_or_dataset_dir):
    """
    get the root directory of the n5 structure
        containing a group or dataset directory
    """
    def is_n5_path(p):
        with (p / "attributes.json").open() as f:
            j = json.load(f)
        if "n5" in j:
            return True
        return False
    group_or_dataset_path = pathlib.Path(group_or_dataset_dir)
    for parent in iter_parents(group_or_dataset_path):
        if is_n5_path(parent):
            return str(parent)


    
class Preprocess(ags.ArgSchemaParser):
        
    def run(self):
        chunk_volume(self.args['input_dir'], self.args['output_dir'],
            chunk_size=self.args['chunk_size'], overlap = self.args['overlap'])
   
        
if __name__ == "__main__":
    mod = Preprocess(schema_type=PreprocessParameters)
    mod.run()      


from ac_segmentation.neurotorch.datasets.dataset import (AlignedVolume, Array, PooledVolume)
from ac_segmentation.neurotorch.datasets.filetypes import (TiffVolume, Hdf5Volume)
from ac_segmentation.neurotorch.datasets.datatypes import (BoundingBox, Vector)
from abc import (ABC, abstractmethod)
import json
import os


class Spec(ABC):
    """
    An abstract class for specifying a volume dataset structure
    """
    def openVolume(self, volume_spec):
        """
        Opens a volume from a volume specification

        :param volume_spec: A dictionary specifying the volume's parameters

        :return: The volume corresponding to the volume dataset
        """
        try:
            filename = os.path.abspath(volume_spec["filename"])

            if filename.endswith(".tif"):
                edges = volume_spec["bounding_box"]
                bounding_box = BoundingBox(Vector(*edges[0]),
                                           Vector(*edges[1]))
                volume = TiffVolume(filename, bounding_box)

                return volume

            elif filename.endswith(".hdf5"):
                pooled_volume = PooledVolume()
                for dataset in volume_spec["datasets"]:
                    edges = dataset["bounding_box"]
                    bounding_box = BoundingBox(Vector(*edges[0]),
                                               Vector(*edges[1]))
                    volume = Hdf5Volume(filename, dataset, bounding_box)
                    pooled_volume.add(volume)

                return pooled_volume

            else:
                error_string = "{} is an unsupported filetype".format(volume_type)
                raise ValueError(error_string)

        except KeyError:
            error_string = "given volume_spec is corrupt"
            raise ValueError(error_string)

    def create(self, spec, stack_size=33):
        """
        Creates a pooled volume from a volume dataset specification

        :param spec: An array of dictionaries specifying the volume's parameters
        :param stack_size: The maximum number of open volumes

        :return: The pooled volume of the volume dataset
        """
        pooled_volume = PooledVolume(stack_size=stack_size)

        for item in spec:
            volume = self.openVolume(item)
            pooled_volume.add(volume)

        return pooled_volume

    def open(self, spec_filename):
        """
        Opens a pooled volume from a volume dataset specification file

        :param spec_filename: The filename of the volume dataset specification
        :return: The pooled volume of the volume dataset
        """
        spec = self.parse(spec_filename)

        cwd = os.getcwd()
        os.chdir(os.path.dirname(spec_filename))
        pooled_volume = self.create(spec)
        os.chdir(cwd)

        return pooled_volume

    @abstractmethod
    def parse(self, spec_filename):
        """
        Parses a specification file

        :param spec_filename: The filename of the volume dataset specification

        :return: An array of dictionaries specifying the volume's parameters
        """
        pass


class JsonSpec(Spec):
    def parse(self, spec_filename):
        """
        Parses a JSON specification file

        :param spec_filename: The filename of the JSON specification

        :return: An array of dictionaries specifying the volume's parameters
        """
        if not spec_filename.endswith(".json"):
            error_string = "{} is not a JSON file".format(spec_filename)
            raise ValueError(error_string)

        with open(spec_filename, 'r') as f:
            spec = json.load(f)

            return spec



import torch
from torch.utils.data import Dataset as _Dataset
import numpy as np
from abc import abstractmethod
from numbers import Number
from numpy import ndarray
from scipy.spatial import KDTree
from functools import reduce
import math
import scipy
import tensorstore as ts
from ac_segmentation.neurotorch.datasets.datatypes import BoundingBox, Vector
import h5py
import tifffile as tif
import os



class Data:
    """
    An encapsulating object for communicating volumetric data
    """
    def __init__(self, array: ndarray, bounding_box: BoundingBox):
        """
        Initializes a data packet from an Numpy array and its bounding box

        :param array: A Numpy array containing the data packet's contents
        :param bounding_box: A bounding box specifying the data packet's
location in 3D-space
        """
        self._setBoundingBox(bounding_box)
        self._setArray(array)

    def getBoundingBox(self) -> BoundingBox:
        """
        Returns the bounding box specifying the data packet's location in
3D-space

        :return: The bounding box of the data packet's location
        """
        return self.bounding_box

    def _setBoundingBox(self, bounding_box: BoundingBox):
        """
        Set the bounding box specifying the data packet's location in 3D-space

        :param bounding_box: The bounding box of the data packet's location
        """
        if not isinstance(bounding_box, BoundingBox):
            raise ValueError("bounding_box must have type BoundingBox")

        self.bounding_box = bounding_box

    def getArray(self) -> ndarray:
        """
        Retrieves the data packet's contents

        :return: An Numpy ndarray in row-major order (Z, Y, X)
        """
        return self.array

    def _setArray(self, array):
        """
        Sets the data packet's contents

        :param array: An Numpy ndarray in row-major order (Z, Y, X)
        """
        self.array = array

    def getSize(self) -> Vector:
        """
        Retrieves the size of the data packet

        :return: A vector containing the data packet's size
        """
        return self.getBoundingBox().getSize()

    def __add__(self, other):
        if not isinstance(other, Data):
            raise ValueError("other must have type Data")
        if self.getBoundingBox() != other.getBoundingBox():
            raise ValueError("other must have the same bounding box")

        return Data(self.getArray() + other.getArray(),
                    self.getBoundingBox())

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        return (self * -1)

    def __mul__(self, other):
        if not isinstance(other, Number):
            error_string = "other must be a number instead it is a {}"
            error_string = error_string.format(type(other))
            raise ValueError(error_string)

        return Data(self.getArray() * other,
                    self.getBoundingBox())

    def __div__(self, other):
        if not isinstance(other, Number):
            error_string = "other must be a number instead it is a {}"
            error_string = error_string.format(type(other))
            raise ValueError(error_string)

        return (self * (1/other))

        
        
class Array:
    """
    A dataset containing a 3D volumetric array
    """
    def __init__(self, array: np.ndarray, bounding_box: BoundingBox=None,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(128, 128, 32)),
                 stride: Vector=Vector(64, 64, 16)):
        """
        Initializes a volume with a bounding box and iteration parameters

        :param array: A 3D Numpy array
        :param bounding_box: The bounding box encompassing the volume
        :param iteration_size: The bounding box of each data sample in the
dataset iterable
        :param stride: The stride displacement of each data sample in the
dataset iterable. The displacement proceeds first from X then to Y then to Z.
        """
        if isinstance(array, np.ndarray):
            self._setArray(array)
        elif isinstance(array, BoundingBox):
            self.createArray(array)
        else:
            raise ValueError("array must be an ndarray or a BoundingBox")

        self.setBoundingBox(bounding_box)
        self.setIteration(iteration_size=iteration_size,
                          stride=stride)
        super().__init__()

    def get(self, bounding_box: BoundingBox) -> Data:
        """
        Requests a data sample from the volume. If the bounding box does
not exist, then the method raises a ValueError.

        :param bounding_box: The bounding box of the request data sample
        :return: The data sample requested
        """
        if bounding_box.isDisjoint(self.getBoundingBox()):
            error_string = ("Bounding box must be inside dataset " +
                            "dimensions instead bounding box is {} while " +
                            "the dataset dimensions are {}")
            error_string = error_string.format(bounding_box,
                                               self.getBoundingBox())
            raise ValueError(error_string)

        sub_bounding_box = bounding_box.intersect(self.getBoundingBox())
        array = self.getArray(sub_bounding_box)

        before_pad = bounding_box.getEdges()[0] - sub_bounding_box.getEdges()[0]
        after_pad = bounding_box.getEdges()[1] - sub_bounding_box.getEdges()[1]

        if before_pad != Vector(0, 0, 0) or after_pad != Vector(0, 0, 0):
            pad_size = tuple(zip(before_pad.getNumpyDim(),
                                 after_pad.getNumpyDim()))
            array = np.pad(array, pad_width=pad_size, mode="constant")

        return Data(array, bounding_box)

    def set(self, data: Data):
        """
        Sets a section of the volume within the provided bounding box with the
given data.

        :param data: The data packet to set the volume
        """
        data_bounding_box = data.getBoundingBox()
        data_array = data.getArray()

        if not data_bounding_box.isSubset(self.getBoundingBox()):
            raise ValueError("The bounding box must be a subset of the "
                             " volume")

        data_edge1, data_edge2 = data_bounding_box.getEdges()
        array_edge1, array_edge2 = self.getBoundingBox().getEdges()

        edge1 = data_edge1 - array_edge1
        edge2 = data_edge2 - array_edge1

        x1, y1, z1 = edge1.getComponents()
        x2, y2, z2 = (min(s, c) for s, c in zip(self.array.shape[::-1], edge2.getComponents()))

        self.array[z1:z2, y1:y2, x1:x2] = data_array[:z2 - z1, :y2 - y1, :x2 - x1]

    def blend(self, data: Data):
        """
        Blends a section of the volume within the provided bounding box with
the given data by taking the elementwise maximum value.

        :param data: The data packet to blend into the volume
        """
        array = self.get(data.getBoundingBox()).getArray()
        array = np.maximum(array, data.getArray())

        result = Data(array, data.getBoundingBox())

        self.set(result)

    def getArray(self, bounding_box: BoundingBox=None) -> np.ndarray:
        """
        Retrieves the array contents of the volume. If a bounding box is
provided, the subsection is returned.

        :param bounding_box: The bounding box of a subsection of the volume.
If the bounding box is outside of the volume, a ValueError is raised.
        """
        if bounding_box is None:
            return self.array

        else:
            if not bounding_box.isSubset(self.getBoundingBox()):
                raise ValueError("The bounding box must be a subset" +
                                 " of the volume")

            centered_bounding_box = bounding_box - self.getBoundingBox().getEdges()[0]
            edge1, edge2 = centered_bounding_box.getEdges()
            x1, y1, z1 = edge1.getComponents()
            x2, y2, z2 = edge2.getComponents()

            return self.array[z1:z2, y1:y2, x1:x2]

    def _setArray(self, array):
        self.array = array

    def getBoundingBox(self) -> BoundingBox:
        """
        Retrieves the bounding box of the volume

        :return: The bounding box of the volume
        """
        return self.bounding_box

    def setBoundingBox(self, bounding_box: BoundingBox=None,
                       displacement: Vector=None):
        """
        Sets the bounding box of the volume. By default, it sets the bounding
box to the volume size

        :param bounding_box: The bounding box of the volume
        :param displacement: The displacement of the bounding box from the
origin
        """
        if bounding_box is None:
            self.bounding_box = BoundingBox(Vector(0, 0, 0),
                                            Vector(*self.getArray().shape[::-1]))
        else:
            self.bounding_box = bounding_box

        if displacement is not None:
            self.bounding_box = self.bounding_box + displacement

    def setIteration(self, iteration_size, stride):
        """
        TODO: hack -- allow reading chunks beyond bounds by implementing a real ceiling
        Sets the parameters for iterating through the dataset

        :param iteration_size: The size of each data sample in the volume
        :param stride: The displacement of each iteration
        """
        if not isinstance(iteration_size, BoundingBox):
            error_string = (
                "iteration_size must have type BoundingBox"
                " instead it has type {}")
            error_string = error_string.format(type(iteration_size))
            raise ValueError(error_string)

        if not isinstance(stride, Vector):
            raise ValueError("stride must have type Vector")

        if not iteration_size.isSubset(BoundingBox(
                Vector(0, 0, 0),
                self.getBoundingBox().getSize())):
            raise ValueError("iteration_size must be smaller than volume size")

        self.setIterationSize(iteration_size)
        self.setStride(stride)

        def ceil(x):
            return math.ceil(x)
            # return int(round(x))

        self.element_vec = Vector(*map(
            lambda L, l, s: ceil((L- l) / s + 1),
            self.getBoundingBox().getSize().getComponents(),
            self.iteration_size.getSize().getComponents(),
            self.stride.getComponents()))

        self.index = 0

    def setIterationSize(self, iteration_size):
        self.iteration_size = BoundingBox(Vector(0, 0, 0),
                                          iteration_size.getSize())

    def setStride(self, stride):
        self.stride = stride

    def getIterationSize(self):
        return self.iteration_size

    def getStride(self):
        return self.stride

    def __len__(self):
        return self.element_vec[0]*self.element_vec[1]*self.element_vec[2]

    def __getitem__(self, idx):
        bounding_box = self._indexToBoundingBox(idx)
        result = self.get(bounding_box)

        return result

    def _indexToBoundingBox(self, idx):
        if idx >= len(self):
            self.index = 0
            raise StopIteration

        element_vec = np.unravel_index(idx,
                                       shape=self.element_vec.getComponents())

        element_vec = Vector(*element_vec)
        bounding_box = self.iteration_size+self.stride*element_vec

        return bounding_box

    def __enter__(self):
        pass

    def __exit__(self):
        pass
        
        
class TSArray(Array):
    
    def __init__(self, tensor: ts.TensorStore, bounding_box: BoundingBox=None,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(32, 32, 32)),
                 stride: Vector=Vector(64, 64, 64), prob_map: bool=False):
        """
        Initializes a volume with a bounding box and iteration parameters

        :param tensor: A 3D tensorstore object
        :param bounding_box: The bounding box encompassing the volume
        :param iteration_size: The bounding box of each data sample in the
        dataset iterable
        :param stride: The stride displacement of each data sample in the
        dataset iterable. The displacement proceeds first from X then to Y then to Z.
        """
        if isinstance(tensor, ts.TensorStore):
            self.tensor = tensor
            self.shift = [list(tensor.domain[0])[0], list(tensor.domain[1])[0], list(tensor.domain[2])[0]]
            self.dtype = str(self.tensor.dtype).split('"')[1]
            pass
        else:
            raise ValueError("Array must be a tensorstore object")

        self.setBoundingBox(bounding_box)
        self.setIteration(iteration_size=iteration_size,
                          stride=stride)
        self.prob_map = prob_map
        
    def set(self, data):
        """Allow dropping data off of improperly shaped arrays"""
        data_bounding_box = data.getBoundingBox()
        data_array = data.getArray()
            
        #get edge components 
        data_edge1, data_edge2 = data_bounding_box.getEdges()
        array_edge1, array_edge2 = self.getBoundingBox().getEdges()
        edge1 = data_edge1 - array_edge1
        edge2 = data_edge2 - array_edge1
        x1, y1, z1 = edge1.getComponents()
        x2, y2, z2 = (min(s, c) for s, c in zip(self.tensor.shape[0:3][::-1], edge2.getComponents()))

        return self.tensor[z1:z2, y1:y2, x1:x2].write(data_array[:z2-z1, :y2-y1, :x2-x1].astype(self.dtype))

    def setBoundingBox(self, bounding_box: BoundingBox=None,
                       displacement: Vector=None):
        """
        Sets the bounding box of the volume. By default, it sets the bounding
        box to the volume size

        :param bounding_box: The bounding box of the volume
        :param displacement: The displacement of the bounding box from the
        origin
        """
        if bounding_box is None:
            x1,y1,z1 = (list(self.tensor.domain[0])[0], list(self.tensor.domain[1])[0], list(self.tensor.domain[2])[0])
            x2,y2,z2  = (list(self.tensor.domain[0])[-1]+1, list(self.tensor.domain[1])[-1]+1, list(self.tensor.domain[2])[-1]+1)
            self.bounding_box = BoundingBox(Vector(z1,y1,x1),Vector(z2,y2,x2))
            
        else:
            self.bounding_box = bounding_box

        if displacement is not None:
            self.bounding_box = self.bounding_box + displacement

    def _indexToBoundingBox(self, idx):
        if idx >= len(self):
            self.index = 0
            raise StopIteration

        element_vec = np.unravel_index(idx,
                                       shape=self.element_vec.getComponents())
        element_vec = Vector(*element_vec)
        bounding_box = self.iteration_size+self.stride*element_vec

        if self.shift != [0,0,0]:
            bounding_box.edge1 += Vector(*self.shift[::-1])
            bounding_box.edge2 += Vector(*self.shift[::-1])

        return bounding_box
    
    def __getitem__(self, idx):
        bounding_box = self._indexToBoundingBox(idx)
        result = self.get(bounding_box)
        
        return result

    def _indexToTensor(self, idx):
        if idx >= len(self):
            self.index = 0
            raise StopIteration

        element_vec = np.unravel_index(idx,
                                       shape=self.element_vec.getComponents())

        element_vec = Vector(*element_vec)
        bounding_box = self.iteration_size+self.stride*element_vec
        edge1, edge2 = bounding_box.getEdges()
        x1, y1, z1 = edge1.getComponents()
        x2, y2, z2 = edge2.getComponents()
        
        return self.tensor[z1:z2,y1:y2,x1:x2]

    def get(self, bounding_box: BoundingBox) -> Data:
        """
        Requests a data sample from the volume. If the bounding box does
        not exist, then the method raises a ValueError.

        :param bounding_box: The bounding box of the request data sample
        :return: The data sample requested
        """
        if bounding_box.isDisjoint(self.getBoundingBox()):
            error_string = ("Bounding box must be inside dataset " +
                            "dimensions instead bounding box is {} while " +
                            "the dataset dimensions are {}")
            error_string = error_string.format(bounding_box,
                                               self.getBoundingBox())
            raise ValueError(error_string)

        sub_bounding_box = bounding_box.intersect(self.getBoundingBox())
        array = self.getArray(sub_bounding_box)

        before_pad = bounding_box.getEdges()[0] - sub_bounding_box.getEdges()[0]
        after_pad = bounding_box.getEdges()[1] - sub_bounding_box.getEdges()[1]

        pad_size = ((0, 0), (0, 0), (0, 0))
        if before_pad != Vector(0, 0, 0) or after_pad != Vector(0, 0, 0):
            pad_size = tuple(zip(before_pad.getNumpyDim(),
                                 after_pad.getNumpyDim()))

        out_data = Data(array, bounding_box)
        out_data.pad_size = pad_size
        
        return out_data

    def getArray(self, bounding_box: BoundingBox=None) -> np.ndarray:
        """
        Retrieves the array contents of the volume. If a bounding box is
        provided, the subsection is returned.

        :param bounding_box: The bounding box of a subsection of the volume.
        If the bounding box is outside of the volume, a ValueError is raised.
        """
        if bounding_box is None:
            return self.tensor[z1:z2, y1:y2, x1:x2].read()

        else:
            if not bounding_box.isSubset(self.getBoundingBox()):
                raise ValueError("The bounding box must be a subset" +
                                 " of the volume")
                
            centered_bounding_box = bounding_box - self.getBoundingBox().getEdges()[0]
            edge1, edge2 = centered_bounding_box.getEdges()
            x1, y1, z1 = [sum(x) for x in zip(edge1.getComponents(),self.shift[::-1])]
            x2, y2, z2 = [sum(x) for x in zip(edge2.getComponents(),self.shift[::-1])]

            return self.tensor[z1:z2, y1:y2, x1:x2].read()

    def blend(self, data: Data, shift):
        """
        Blends a section of the volume within the provided bounding box with
        the given data by taking the elementwise maximum value.

        :param data: The data packet to blend into the volume
        """

        v1 = data.bounding_box.edge1 - Vector(*shift[::-1])
        v2 = data.bounding_box.edge2 - Vector(*shift[::-1])
        data_bb = BoundingBox(v1,v2)
        
        array = self.get(data_bb).getArray().result()
        data_array = data.getArray()
        
        if self.prob_map == True:
            data_array = torch.special.expit(torch.from_numpy(data_array)).numpy()
        x,y,z = array.shape
        array = np.maximum(array, data_array[:x,:y,:z])
        result = Data(array, data_bb)
        write = self.set(result)
        
        return write
        
    def out_hdf5(self, dir):
        hf = h5py.File(dir + 'Seg_Array.h5', 'w')
        for i in range(self.tensor.shape[0]):
            hf.create_dataset('z' + str(i), data=self.tensor[i,:,:].read().result())
        hf.close()

    def out_tiff(self, dir):
        path = os.path.join(dir, 'Seg_Tiffs/')
        os.mkdir(path) 
        for i in range(self.tensor.shape[0]):
            tif.imwrite(path + '%03d.tif'%i, self.tensor[i,:,:].read().result())

    def setMask(self, mask):
        if isinstance(mask, ts.TensorStore):
            self.mask = TSArray(mask, stride=self.stride, iteration_size=self.iteration_size)
        elif isinstance(mask, np.ndarray):
            self.mask = Array(mask, stride=self.stride, iteration_size=self.iteration_size)
        else:
            raise ValueError("Mask must be a tensorstore object or numpy array")

    def setLiThreshold(self, li_thresh):
        if isinstance(li_thresh, int) or isinstance(li_thresh, float):
            self.li_thresh = li_thresh
        else:
            raise ValueError("Li threshold must be an integer or float")
            
    def set_Rescale_Perc(self, percentile=(70,100)):
        self.rescale_perc = percentile
            


class TorchVolume(_Dataset):
    def __init__(self, volume):
        self.setVolume(volume)
        super().__init__()

    def __len__(self):
        return len(self.getVolume())

    def __getitem__(self, idx):
        if isinstance(self.getVolume(), AlignedVolume):
            data_list = [self.toTorch(data) for data in self.getVolume()[idx]]
            return data_list
        else:
            return self.getVolume()[idx].getArray()

    def toTorch(self, data):
        torch_data = data.getArray().astype(np.float)
        torch_data = torch_data.reshape(1, *torch_data.shape)
        return torch_data

    def setVolume(self, volume):
        self.volume = volume

    def getVolume(self):
        return self.volume


class Volume:
    """
    An interface for creating volumes
    """
    def __init__(self, bounding_box: BoundingBox=None,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(128, 128, 32)),
                 stride: Vector=Vector(64, 64, 16)):
        self.setBoundingBox(bounding_box)
        self.setIteration(iteration_size, stride)
        self.valid_data = None

    def setArray(self, array: Array):
        self.array = array

    def getArray(self) -> Array:
        return self.array

    def request(self, bounding_box):
        return self.getArray().get(bounding_box)

    def set(self, data: Data):
        self.getArray().set(data)

    def blend(self, data: Data):
        self.getArray().blend(data)

    def setIteration(self, iteration_size: BoundingBox, stride: Vector):
        if not isinstance(iteration_size, BoundingBox):
            error_string = ("iteration_size must have type BoundingBox"
                            + " instead it has type {}")
            error_string = error_string.format(type(iteration_size))
            raise ValueError(error_string)

        if not isinstance(stride, Vector):
            raise ValueError("stride must have type Vector")

        if not iteration_size.isSubset(BoundingBox(Vector(0, 0, 0),
                                                   self.getBoundingBox().getSize())):
            raise ValueError("iteration_size must be smaller than volume size " +
                             "instead the iteration size is {} ".format(iteration_size.getSize()) +
                             "and the volume size is {}".format(self.getBoundingBox().getSize()))

        self.setIterationSize(iteration_size)
        self.setStride(stride)

        def ceil(x):
            return int(round(x))

        self.element_vec = Vector(*map(lambda L, l, s: ceil((L-l)/s+1),
                                       self.getBoundingBox().getSize().getComponents(),
                                       self.iteration_size.getSize().getComponents(),
                                       self.stride.getComponents()))

        self.index = 0

    def setBoundingBox(self, bounding_box):
        if not isinstance(bounding_box, BoundingBox):
            raise ValueError("bounding_box must have type BoundingBox " +
                             "instead it has type {}".format(type(bounding_box)))
        self.bounding_box = bounding_box

    def getBoundingBox(self):
        return self.bounding_box

    def setIterationSize(self, iteration_size):
        if not isinstance(iteration_size, BoundingBox):
            raise ValueError("iteration_size must have type BoundingBox")
        self.iteration_size = iteration_size

    def getIterationSize(self):
        return self.iteration_size

    def setStride(self, stride):
        if not isinstance(stride, Vector):
            raise ValueError("stride must have type Vector")
        self.stride = stride

    def getStride(self):
        return self.stride

    @abstractmethod
    def loadArray(self):
        pass

    @abstractmethod
    def unloadArray(self):
        pass

    @abstractmethod
    def get(self, bounding_box: BoundingBox) -> Data:
        """
        Requests a data sample from the dataset. If the bounding box does
not exist, then the method raises a ValueError.

        :param bounding_box: The bounding box of the request data sample
        :return: The data sample requested
        """
        pass

    @abstractmethod
    def set(self, data: Data):
        """
        Sets a section of the dataset within the provided bounding box with the
given data.

        :param data: The data packet to set the volume
        """
        pass

    @abstractmethod
    def __enter__(self):
        """
        Loads the dataset into memory
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Unloads the dataset from memory
        """
        pass

    def __len__(self) -> int:
        """
        Returns the length of the dataset

        :return: The dataset length
        """
        return self.element_vec[0]*self.element_vec[1]*self.element_vec[2]

    def __getitem__(self, idx: int):
        """
        Returns the data sample at index idx from the dataset

        :param idx: The index of the data sample
        """
        return self.getArray()[idx]

    def __iter__(self):
        """
        Returns an iterable of the dataset

        :return: The iterable of the dataset
        """
        self.index = 0
        return self

    def __next__(self):
        """
        Retrieves the next data sample from the dataset
        :return: The next data sample
        """
        if self.index < len(self):
            result = self.__getitem__(self.index)
            self.index += 1
            return result
        else:
            raise StopIteration

    def getValidData(self):
        if self.valid_data is not None:
            self.valid_data = []
            for i in range(len(self)):
                if not (self[i].getArray() == 0).all():
                    self.valid_data.append(i)

        return self.valid_data


class AlignedVolume(Volume):
    def __init__(self, volumes, iteration_size=None, stride=None):
        if iteration_size is None:
            iteration_size = volumes[0].getIterationSize()
        if stride is None:
            stride = volumes[0].getStride()
        self.setVolumes(volumes)
        self.setIteration(iteration_size, stride)
        self.valid_data = None

    def getBoundingBox(self):
        return self.getVolumes()[0].getBoundingBox()

    def setVolumes(self, volumes):
        self.volumes = volumes

    def addVolume(self, volume):
        self.volumes.append(volume)

    def getVolumes(self):
        return self.volumes

    def setIteration(self, iteration_size, stride):
        for volume in self.getVolumes():
            volume.setIteration(iteration_size, stride)

    def get(self, bounding_box):
        result = [volume.get(bounding_box)
                  for volume in self.getVolumes()]
        return result

    def set(self, array, bounding_box):
        pass

    def __len__(self):
        return len(self.getVolumes()[0])

    def __getitem__(self, idx):
        result = [volume[idx] for volume in self.getVolumes()]
        return result

    def getValidData(self):
        if self.valid_data is None:
            self.valid_data = []
            for i in range(len(self)):
                if not (self.getVolumes()[1][i].getArray() == 0).all():
                    self.valid_data.append(i)

        return self.valid_data

    def _indexToBoundingBox(self, idx):
        bounding_box = self.getVolumes()[0]._indexToBoundingBox(idx)

        return bounding_box


class PooledVolume(Volume):
    def __init__(self, volumes=None, stack_size: int=5,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(128, 128, 32)),
                 stride: Vector=Vector(64, 64, 16)):
        if volumes is not None:
            self.volumes = volumes
            self.volumes_changed = True
        else:
            self.volumes = []
            self.volumes_changed = False

        self.volume_list = []
        self.setStack(stack_size)

        self.setIteration(iteration_size, stride)

        self.valid_data = None

    def setStack(self, stack_size: int=15):
        self.stack = []
        self.stack_size = stack_size

    def _pushStack(self, index, volume):
        if len(self.stack) >= self.stack_size:
            self.stack[0][1].__exit__(None, None, None)
            self.stack.pop(0)

        pos = len(self.stack)
        self.stack.insert(pos, (index, volume.__enter__()))

        return pos

    def _rebuildIndexes(self):
        edge1_list = [volume.getBoundingBox().getEdges()[0].getComponents()
                      for volume in self.volumes]

        self.edge1_list = KDTree(edge1_list)

        self.__len__()

        self.volumes_changed = False

    def _queryBoundingBox(self, bounding_box: BoundingBox) -> Volume:
        if self.volumes_changed:
            self._rebuildIndexes()

        edge1 = [bounding_box.getEdges()[0].getComponents()]
        distances, indexes  = self.edge1_list.query(edge1, k=8)
        indexes = [index for index, dist in zip(indexes[0], distances[0])
                   if dist < float('Inf')]
        indexes = filter(lambda index: not bounding_box.isDisjoint(self.volumes[index].getBoundingBox()),
                         indexes)
        if not indexes:
            raise IndexError("bounding_box is not present in any indexes")

        return list(indexes)

    def add(self, volume: Volume):
        self.volumes_changed = True
        self.volumes.append(volume)

    def get(self, bounding_box: BoundingBox) -> Data:
        indexes = self._queryBoundingBox(bounding_box)

        data = []

        stack_volumes = [volume for i, volume in self.stack if i in indexes]
        stack_disjoint = list(set(indexes) - set([i for i, v in self.stack]))

        for volume in stack_volumes:
            sub_bbox = bounding_box.intersect(volume.getBoundingBox())
            data.append(volume.get(sub_bbox))

        for index in stack_disjoint:
            volume = self.volumes[index]
            i = self._pushStack(index, volume)

            sub_bbox = bounding_box.intersect(volume.getBoundingBox())
            data.append(volume.get(sub_bbox))

        shape = bounding_box.getNumpyDim()
        array = Array(np.zeros(shape).astype(np.uint16),
                        bounding_box=bounding_box,
                        iteration_size=BoundingBox(Vector(0, 0, 0),
                                                    bounding_box.getSize()),
                        stride=bounding_box.getSize())
        [array.set(item) for item in data]
        return Data(array.getArray(), bounding_box)

    def set(self, data: Data):
        indexes = self._queryBoundingBox(data.getBoundingBox())

        data = []
        for index in indexes:
            for stack_index, stack_volume in self.stack:
                if stack_index == index:
                    stack_volume.set(data)
                else:
                    volume = self.volume_list[index].__enter__()
                    self._pushStack(index, volume)

                    volume.set(data)

    def __exit__(self, exc_type, exc_value, traceback):
        for index, volume in self.stack:
            volume.__exit__()

    def __len__(self) -> int:
        if self.volumes_changed:
            self.volume_index = [0]
            for volume in self.volumes:
                self.volume_index.append(self.volume_index[-1] + len(volume))
            self.length = self.volume_index[-1]

        return self.length

    def __getitem__(self, idx: int) -> Data:
        bounding_box = self._indexToBoundingBox(idx)
        result = self.get(bounding_box)

        return result

    def _indexToBoundingBox(self, idx: int) -> BoundingBox:
        if self.volumes_changed:
            len(self)

        if idx >= len(self):
            self.index = 0
            raise StopIteration

        index = max(filter(lambda index: self.volume_index[index] <= idx,
                           range(len(self.volume_index))))
        volume = self.volumes[index]
        _idx = idx-self.volume_index[index]

        element_vec = np.unravel_index(_idx,
                                       shape=volume.element_vec.getComponents())

        element_vec = Vector(*element_vec)
        bounding_box = volume.iteration_size+volume.stride*element_vec \
                       + volume.getBoundingBox().getEdges()[0]

        return bounding_box

    def setIteration(self, iteration_size: BoundingBox, stride: Vector):
        for volume in self.volume_list:
            volume.setIteration(iteration_size, stride)

        self.setIterationSize(iteration_size)
        self.setStride(stride)

    def getValidData(self):
        if self.valid_data is None:
            self.valid_data = []
            for i in range(len(self)):
                if not (self[i].getArray() == 0).all():
                    self.valid_data.append(i)

        return self.valid_data



import pytest
import os
import glob
from ac_segmentation.process.segment_and_skeletonize_strip import run


def test_seg_skel(zarr_file):
    path = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + os.sep + os.pardir)
    weights = glob.glob(f"{path}/**/{'best.ckpt'}", recursive=True)[0]
    run(weights_file=weights, 
        input_zarr=zarr_file+"test_zarr.zarr", 
        probability_output_path=os.path.join(zarr_file, "output_seg.npy.gz"),
            skeleton_output_path=os.path.join(zarr_file, "output_skel.swcs.tar.gz"), zarr_level=0, probability_threshold=0.05,
            label_size_threshold=80, filter_max_intensity=30000)




from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
import random
import numpy as np
from scipy.ndimage.filters import convolve


class Duplicate(Augmentation):
    def __init__(self, volume, max_slices=20, **kwargs):
        self.setMaxSlices(max_slices)
        super().__init__(volume, **kwargs)

    def augment(self, bounding_box):
        slices = self.getSlices()
        end = bounding_box.getSize().getComponents()[0]
        location = random.randrange(end-slices)

        raw_data, label_data = self.getParent().get(bounding_box)
        augmented_raw, augmented_label = self.duplication(raw_data, label_data,
                                                          location=location,
                                                          slices=slices)

        return (augmented_raw, augmented_label)

    def setMaxSlices(self, max_slices):
        self.max_slices = max_slices

    def getMaxSlices(self):
        return self.max_slices

    def getSlices(self):
        return random.randrange(2, self.getMaxSlices())

    def duplication(self, raw_data, label_data, location=20, slices=3,
                    axis=0):
        raw = raw_data.getArray()
        distorted_raw = raw.copy()

        noise = raw[:, :, location:location+slices]
        noise = noise - convolve(noise, weights=np.full((3, 3, 3), 1.0/27))

        duplicate_slices = np.repeat(raw[:, :, location].reshape(raw.shape[0],
                                                                 raw.shape[1],
                                                                 1),
                                     slices, axis=2)
        duplicate_slices += noise
        distorted_raw[:, :, location:location+slices] = duplicate_slices

        augmented_raw_data = Data(distorted_raw, raw_data.getBoundingBox())

        return augmented_raw_data, label_data



import os
import numpy as np
import pandas as pd
import argschema as ags
from operator import add
from collections import deque, defaultdict
from joblib import dump, load
import copy
from sklearn.neighbors import KDTree
from scipy.spatial.distance import euclidean
from ac_segmentation.deprecated.swc_morphology.classes import Morphology
from ac_segmentation.deprecated.swc_morphology.readwrite_swc import morphology_from_swc, morphology_to_swc

class ReconnectParameters(ags.ArgSchema):
    input_file = ags.fields.InputFile(required=True, description='Input file')
    output_dir = ags.fields.OutputDir(required=True, description='Output directory')
    model_dir = ags.fields.InputDir(required=True, description = 'Model directory')
    pxl_xyz = ags.fields.NumpyArray(dtype=float, required=False, 
                           default=[0.812,0.812,0.704], description='pxl size in um')

def reconnect(infile, swc_outdir, modeldir, xyz_pxl): 
    if not os.path.isdir(swc_outdir):
        os.mkdir(swc_outdir)
    
    # Prune short branches 
    swc_prune(infile, os.path.join(swc_outdir, 'pruned.swc'),
                pruning_threshold = 15)
    
    # Split branches and save all segments as a single file
    swc_split_branches(os.path.join(swc_outdir, 'pruned.swc'),
                        os.path.join(swc_outdir, 'segments.swc'))
    sort_swc(os.path.join(swc_outdir, 'segments.swc'),
             os.path.join(swc_outdir, 'segments.swc'))  
    
    # # Upsample swc using vaa3d
    
    # Reconnect segments
    input_file = os.path.join(swc_outdir, 'segments_resampled.swc')
    # Load scaler
    scaler = load(os.path.join(modeldir, 'scaler.joblib'))
    # Load classifier model
    clf = load(os.path.join(modeldir, 'LR_1.joblib')) 
    model_pxl = np.array([0.207,0.207,0.6]) # model pxl
    xyz_pxl = xyz_pxl*np.mean(model_pxl/xyz_pxl)
    max_iter = 3
    thresh_list = [0.5, 0.5, 0.5]
    for num_iter in range(1,max_iter+1):
        threshold = thresh_list[num_iter-1]     
        print('num_iter', num_iter, 'thresh', threshold, 'input_file', input_file)
        
        # Find pairs
        pair_data_iter = find_pairs(input_file, scaler, clf, xyz_pxl, threshold)
                        
        # Remove duplicates
        pair_data_iter = remove_duplicates(pair_data_iter)
        
        # Save pair_data as csv file
        df1 = pd.DataFrame.from_dict(pair_data_iter)
        csv_file = os.path.join(swc_outdir, 'pair_dict_iter%d.csv'%num_iter)
        df1.to_csv(csv_file, index=False)
        
        # Merge segment pairs with prob below thresh
        output_file = os.path.join(swc_outdir, 'connect_iter%d.swc'%num_iter) 
        new_morph = merge_pairs(input_file, output_file, csv_file, threshold)
        input_file = output_file
            
            
def load_swc(filepath):
    "Load swc file as a N X 7 numpy array"
    swc = []
    with open(filepath) as f:
        lines = f.read().split("\n")
        for l in lines:
            if not l.startswith('#'):
                cells = l.split(' ')
                if len(cells) == 7:
                    cells = [float(c) for c in cells]
                    swc.append(cells)
                elif len(cells) == 8:
                    cells = [float(c) for c in cells[0:7]]
                    swc.append(cells)                
    return np.array(swc)                     

def save_swc(filepath, swc):
    with open(filepath, 'w') as f:
        f.write('#id,type,x,y,z,r,pid\n')
        for i in range(swc.shape[0]):
            f.write('%.0f %.0f %.0f %.0f %.0f %.3f %d\n' %tuple(swc[i, :].tolist()))
    
def swc_multi_to_single(dirname, fname):
    file_list = [f for f in os.listdir(dirname) if f.endswith('swc')] 
    file_list.sort()
    trace_list = []
    for f in file_list:
        trace = load_swc(os.path.join(dirname, f))
        trace_list.append(trace)
    offset = 0
    for i, trace in enumerate(trace_list):
        select = np.where(trace[:,-1]!=-1)[0]  
        trace_i = np.copy(trace)
        min_id = np.min(trace_i[:,0])
        trace_i[:,0] = trace_i[:,0] + offset - min_id + 1
        trace_i[select,-1] = trace_i[select, -1] + offset - min_id + 1
        offset = np.max(trace_i[:,0])
        if i == 0:
            trace_new = trace_i
        else:
            trace_new = np.concatenate((trace_new, trace_i))
    save_swc(os.path.join(dirname,fname), trace_new)          

def bfs_tree(st_node,morph):
    "BFS treee traversal, returns nodes in segment and how many"
    queue = deque([st_node])
    nodes_in_segment = []
    seg_len = 0
    while len(queue) > 0:
        seg_len+=1
        current_node = queue.popleft()
        nodes_in_segment.append(current_node)
        for ch_no in morph.get_children(current_node):
            queue.append(ch_no)

    return nodes_in_segment, len(nodes_in_segment)    
    
def swc_prune(infile, outfile, pruning_threshold = 30,**kwargs):
    morph_in = morphology_from_swc(infile)
    nodes_to_remove = set()
    prune_count = 0
    bifur_nodes = [n for n in morph_in.nodes() if len(morph_in.get_children(n)) > 1]
    for bif_node in bifur_nodes:
        children = morph_in.get_children(bif_node)
        for child in children:
            child_remove_nodes, child_seg_length = bfs_tree(child,morph_in)
            if child_seg_length < pruning_threshold:
                prune_count+=1
                [nodes_to_remove.add(n['id']) for n in child_remove_nodes]
                
    keeping_nodes = [n for n in morph_in.nodes() if n['id'] not in nodes_to_remove]

    morph_out = Morphology(
        keeping_nodes,
        node_id_cb=lambda node: node['id'],
        parent_id_cb=lambda node: node['parent'] )
    morphology_to_swc(morph_out, outfile) 
    
def swc_split_branches(infile, outfile):
    morph_in = morphology_from_swc(infile)
    # Find branching nodes
    branching_nodes = morph_in.get_branching_nodes()
    # Split branches
    for node in branching_nodes:
        branching_node_children = morph_in.children_of(node)
        for child in branching_node_children:
            child_id = child['id']        
            morph_in.node_by_id(child_id)['parent'] = -1
    morph_out = Morphology( 
                        morph_in.nodes(),
                        node_id_cb=lambda node: node['id'],
                        parent_id_cb=lambda node: node['parent'])
    # Find all trees remove very short ones
    tree_list = morph_out.get_tree_list()
    new_tree_list = []
    for tree in tree_list:
        if len(tree) > 9:
            new_tree_list.append(tree)
    tree_nodes = [node for tree in new_tree_list for node in tree] 
    morph_out = Morphology(tree_nodes,
                node_id_cb=lambda node: node['id'],
                parent_id_cb=lambda node: node['parent'])
    morphology_to_swc(morph_out, outfile) 

def dfs_labeling(st_node, new_starting_id, modifying_dict, morph):
    ct = 0
    queue = deque([st_node])
    while len(queue) > 0:
        ct+=1
        current_node = queue.popleft()
        modifying_dict[current_node['id']] = new_starting_id
        new_starting_id+=1
        for ch_no in morph.get_children(current_node):
            queue.appendleft(ch_no)
    return ct
    
def sort_swc(infile, outfile):
    morph = morphology_from_swc(infile)
    unordered_swc_info  = {}
    with open(infile,'r') as f:
        for l in f:
            if '#' not in l:
                no_id = int(l.split(' ')[0])
                parent_id = l.split()[-1]
                children_list = morph.get_children(morph.node_by_id(no_id)) 
                unordered_swc_info[no_id] = l 
    new_node_ids = {}
    start_label = 1
    for root in morph.get_roots():
        seg_len = dfs_labeling(root,start_label,new_node_ids,morph)
        start_label += seg_len 

    new_output_dict = {}
    for old_id, old_line in unordered_swc_info.items():
        new_id = new_node_ids[old_id]
        old_parent = int(old_line.split()[-1])
        if old_parent == -1:
            new_parent = -1
        else:
            new_parent = new_node_ids[old_parent] 

        new_line_list = [str(new_id)]+old_line.split(' ')[1:-1] + ['{}\n'.format(new_parent)]
        new_line = " ".join(new_line_list) 
        new_output_dict[new_id] = new_line
        
    with open(outfile,"w") as f2:
        for k in sorted(list(new_output_dict.keys())):
            new_write_line = new_output_dict[k]
            f2.write(new_write_line)    

def distance(node1, node2, pxl_xyz):
    node1_coord = np.array((node1['x'], node1['y'], node1['z']))*pxl_xyz
    node2_coord = np.array((node2['x'], node2['y'], node2['z']))*pxl_xyz
    return euclidean(node1_coord, node2_coord)
    
def get_nodes(segment, end_node, n):
    nodes = [end_node]
    i = 0
    if end_node['parent'] != -1:
        while i < n:
            next_node = [node for node in segment if node['id']==nodes[-1]['parent']]
            if next_node==[]:
                i = n
            else:
                nodes.append(next_node[0])
            i+=1
    else:
        while i < n:
            next_node = [node for node in segment if node['parent']==nodes[-1]['id']]
            if next_node==[]:
                i = n
            elif len(next_node) > 1: 
                i = n
            else:
                nodes.append(next_node[0])
            i+=1
    return nodes 

def calculate_vector(nodes, pxl_xyz):
    nodes_coord = np.vstack([np.array((node['x'], node['y'], node['z']))*pxl_xyz for node in nodes])
    nodes_coord_mean = nodes_coord.mean(axis=0)
    _, _, vv = np.linalg.svd(nodes_coord - nodes_coord_mean)
    vector = vv[0]
    
    # Fix wrong orientation (sign) of vector
    vect_diff = (nodes_coord[-1,:] - nodes_coord[0,:])/np.linalg.norm(nodes_coord[-1,:] - nodes_coord[0,:])
    if np.dot(-vector, vect_diff) > np.dot(vector, vect_diff):
        vector*=-1    
    return vector 

def collinearity(segments, end_nodes, pxl_xyz, num_nodes = [4,49]):
    end_coord = [np.array((node['x'], node['y'], node['z']))*pxl_xyz for node in end_nodes]
    cvect = end_coord[1] - end_coord[0]
    cvect_norm = np.linalg.norm(cvect)
    if cvect_norm != 0:
        cvect = cvect/cvect_norm
        cf = []
        for num in num_nodes:
            vectors = []
            for n in range(len(end_nodes)):
                nodes = get_nodes(segments[n], end_nodes[n], num)
                vect = calculate_vector(nodes, pxl_xyz)
                vectors.append(vect)
            cf.append(np.dot(-vectors[0], cvect))
            cf.append(np.dot(vectors[1], cvect))
    else:
        cf = [np.NaN, np.NaN]
    return cf  

def calculate_features(tree1, end1, tree2, end2, pxl_xyz):
    dist = distance(end1, end2, pxl_xyz)
    treelist = [tree1, tree2]
    endlist = [end1, end2]
    cfactors = collinearity(treelist, endlist, pxl_xyz)
    return np.array([dist] + cfactors)

def find_pairs(fname, sc, cl, pxl_xyz, thresh, radius=range(0,15,5)):
    morph_in = morphology_from_swc(fname)
    leaf_nodes = morph_in.get_leaf_nodes()
    tree_list = morph_in.get_tree_list()
    
    end_node_list = []
    idx_list = []
    for i, tree in enumerate(tree_list):
        root_node = morph_in.get_roots_for_nodes(tree)[0]
        if root_node['type'] != 1:
            idx_list.append(i)
            end_node_list.append(root_node)
        tree_leaf_nodes = [node for node in tree if node in leaf_nodes]
        for node in tree_leaf_nodes:
            idx_list.append(i)
            end_node_list.append(node)
    idx_arr = np.array(idx_list)
    end_node_coord = np.array([np.array((node['x']*pxl_xyz[0], node['y']*pxl_xyz[1],
                            node['z']*pxl_xyz[2])) for node in end_node_list])
    
    kdtree_list = []
    for idx1, tree in enumerate(tree_list):
        s1 = np.where(idx_arr == idx1)[0]
        s2 = np.where(idx_arr != idx1)[0]
        kdtree_list.append(KDTree(end_node_coord[s2], leaf_size=2))
    
    farray = np.zeros((len(end_node_list),len(end_node_list),5))
    n = 0
    num_ends = 5
    pair_data = []
    for idx1, tree in enumerate(tree_list):
        s1 = np.where(idx_arr == idx1)[0]
        s2 = np.where(idx_arr != idx1)[0]
        kdtree = kdtree_list[idx1]
        for s in s1:
            min_dist_idx, dist = kdtree.query_radius(end_node_coord[s].reshape(1,3), radius[-1],
                                              return_distance=True, sort_results=True)
            select = []
            for i in range(len(radius)-1):
                select.append(np.where((dist[0]>radius[i])&(dist[0]<=radius[i+1])))
            for l in select:
                p_arr = s2[min_dist_idx[0][l]]
                idx2_arr = idx_arr[p_arr]
                if len(p_arr) > 0:
                    probs = np.zeros((len(p_arr),))
                    nodes = np.empty((len(p_arr),), dtype=object)
                    features = np.empty((len(p_arr),), dtype=object)
                    for i, p in enumerate(p_arr):
                        if farray[s,p,0] == 0:
                            f = calculate_features(tree_list[idx1], end_node_list[s],
                                                  tree_list[idx2_arr[i]], end_node_list[p], pxl_xyz)
                            farray[s,p,:] = f
                        else:
                            f = farray[s,p,:]
 
                        dist1, min_dist_idx1 = kdtree.query(end_node_coord[s].reshape(1,3), k=num_ends)
                        q_arr = s2[min_dist_idx1[0]]
                        s3 = np.where(q_arr != p)[0][0:num_ends-1]
                        q_arr = q_arr[s3]
                        q_arr = q_arr[:num_ends-1]
                        dist1 = dist1[:,s3]
                        idx3_arr = idx_arr[q_arr]
                        for j, q in enumerate(q_arr):
                            if farray[s,q,0] == 0:
                                f1 = calculate_features(tree_list[idx1], end_node_list[s],
                                              tree_list[idx3_arr[j]], end_node_list[q], pxl_xyz)
                                farray[s,q,:] = f1
                            else:
                                f1 = farray[s,q,:]
                            f = np.concatenate((f,f1))

                        s3 = np.where(idx_arr != idx2_arr[i])[0]
                        kdtree1 = kdtree_list[idx2_arr[i]]
                        dist1, min_dist_idx1 = kdtree1.query(end_node_coord[p].reshape(1,3), k=num_ends)
                        q_arr = s3[min_dist_idx1[0]]
                        s4 = np.where(q_arr != s)[0][0:num_ends-1]
                        q_arr = q_arr[s4]
                        q_arr = q_arr[:num_ends-1]
                        dist1 = dist1[:,s4]
                        idx3_arr = idx_arr[q_arr]
                        for j, q in enumerate(q_arr):
                            if farray[p,q,0] == 0:
                                f1 = calculate_features(tree_list[idx2_arr[i]], end_node_list[p],
                                              tree_list[idx3_arr[j]], end_node_list[q], pxl_xyz)
                                farray[p,q,:] = f1
                            else:
                                f1 = farray[p,q,:]
                            f = np.concatenate((f,f1))
                        features[i] = f
                        x = sc.transform(f.reshape(1,-1))
                        probs[i] = cl.predict_proba(x)[0][1]
                    
                    probs_max = np.max(probs)
                    k_max = np.argmax(probs)
                    if probs_max >= thresh:
                        pair_dict = {}
                        pair_dict['idx'] = n
                        pair_dict['tree1'] = idx1
                        pair_dict['tree2'] = idx2_arr[k_max]
                        pair_dict['nid1'] = end_node_list[s]['id']
                        pair_dict['pid1'] = end_node_list[s]['parent']
                        pair_dict['nid2'] = end_node_list[p_arr[k_max]]['id']
                        pair_dict['pid2'] = end_node_list[p_arr[k_max]]['parent']
                        pair_dict['prob'] = probs_max
                        for i in range(num_ends*2 -1):
                            pair_dict['distance%d'%i] = features[k_max][i*5]
                            pair_dict['cf0_near%d'%i] = features[k_max][i*5+1]
                            pair_dict['cf1_near%d'%i] = features[k_max][i*5+2]
                            pair_dict['cf0_far%d'%i] = features[k_max][i*5+3]
                            pair_dict['cf1_far%d'%i] = features[k_max][i*5+4]
                        pair_data.append(pair_dict)
                        n += 1
                        break
    return pair_data

def remove_duplicates(pair_data): 
    t1 = [pair['tree1'] for pair in pair_data]
    t2 = [pair['tree2'] for pair in pair_data]
    num_tree = max(max(t1), max(t2)) + 1
    dupl_list = []    
    for i in range(num_tree):
        select1 = [pair_data.index(p) for p in pair_data if p['tree1'] == i]
        js = [pair_data[s1]['tree2'] for s1 in select1]
        js_unique = np.unique(js)
        for j in js_unique:
            s2 = np.where(js == j)[0]
            if len(s2) > 1:
                probs = [pair_data[select1[s]]['prob'] for s in s2]
                idx = np.argmax(probs)
                for k, s in enumerate(s2):
                    if k != idx:
                        dupl_list.append(select1[s])
    pair_data1 = [pair_data[i] for i in range(len(pair_data)) if i not in dupl_list]  
    if not pair_data1:
        return
    
    dupl_list = []    
    for i in range(num_tree):
        select1 = [pair_data1.index(p) for p in pair_data1 if p['tree1'] == i]
        for s1 in select1:
            j = pair_data1[s1]['tree2']
            select2 = [pair_data1.index(p) for p in pair_data1 if p['tree1'] == j]
            for s2 in select2:
                k = pair_data1[s2]['tree2']
                if k == i:
                    if pair_data1[s1]['prob'] < pair_data1[s2]['prob']:
                        dupl_list.append(s1)
                    elif pair_data1[s1]['prob'] > pair_data1[s2]['prob']:
                        dupl_list.append(s2)
                    else:    
                        dupl_list.append(max(s1,s2))
    pair_data1 = [pair_data1[i] for i in range(len(pair_data1)) if i not in dupl_list]
    if not pair_data1:
        return
    
    n1 = [pair['nid1'] for pair in pair_data1]
    n2 = [pair['nid2'] for pair in pair_data1]
    num_nid = max(max(n1), max(n2)) + 1
    
    dupl_list = []
    for i in range(num_nid):
        select = []
        select1 = [pair_data1.index(p) for p in pair_data1 if p['nid1'] == i]
        if len(select1) > 0:
            for s1 in select1:
                select.append(s1)
        select2 = [pair_data1.index(p) for p in pair_data1 if p['nid2'] == i]
        if len(select2) > 0:
            for s2 in select2:
                select.append(s2)
        if len(select) > 1:
            probs = [pair_data1[s]['prob'] for s in select]
            idx = np.argmax(probs)
            for k, s in enumerate(select):
                if k != idx:
                    dupl_list.append(s)
    pair_data1 = [pair_data1[i] for i in range(len(pair_data1)) if i not in dupl_list]
    return pair_data1    

def reroot_tree(start_node, tree, morph):
    neighbors_dict = {}
    for node in tree:
        node_neighbors = [] 
        parent = morph.parent_of(node)
        if parent != None:
            node_neighbors.append(parent)
        children = morph.get_children(node)
        for ch in children:
            node_neighbors.append(ch) 
        neighbors_dict[node['id']] = node_neighbors 
    start_node_parent = -1
    
    parent_dict = {}
    parent_dict[start_node['id']] = start_node_parent
    queue = deque([start_node])
    while len(queue) > 0:
        current_node = queue.popleft()
        neighbors = neighbors_dict[current_node['id']]
        for node in neighbors:
            if node['id'] not in parent_dict:
                parent_dict[node['id']] = current_node['id']
                queue.append(node)
    
    new_tree = copy.deepcopy(tree)
    for node in new_tree:
        node['parent'] = parent_dict[node['id']]
    return new_tree  

def connect_trees(nodes, morph_in):
    tree_list = morph_in.get_tree_list()
    
    n = 0
    for i, tree in enumerate(tree_list):
        if nodes[0] in tree:
            tree1 = copy.deepcopy(tree)
            idx1 = i
            n += 1
        if nodes[1] in tree:
            tree2 = copy.deepcopy(tree)
            idx2 = i
            n += 1
        if n == 2:
            break
    if idx1 != idx2:
        i1 = tree1.index(nodes[0])
        if nodes[0]['parent'] == -1:
            leaf_nodes = morph_in.get_leaf_nodes()
            tree_leaf_nodes = [node for node in tree1 if node in leaf_nodes]
            if len(tree_leaf_nodes) == 0:
                print('do not reroot')
            else:    
                tree1 = reroot_tree(tree_leaf_nodes[0], tree1, morph_in)

        i2 = tree2.index(nodes[1])
        if nodes[1]['parent'] != -1:
            tree2 = reroot_tree(nodes[1], tree2, morph_in)
        
        tree2[i2]['parent'] = tree1[i1]['id']
        tree_list[idx1] = tree1
        tree_list[idx2] = tree2
        
        tree_nodes = [node for tree in tree_list for node in tree]
        morph_out = Morphology(tree_nodes,
                    node_id_cb=lambda node: node['id'],
                    parent_id_cb=lambda node: node['parent'])
    else:
        print('loop, do not connect')
        morph_out = morph_in
    return morph_out

def merge_pairs(infile, outfile, csvfile, thresh):
    dfp = pd.read_csv(csvfile)
    pair_data = dfp.to_dict('records')
    morph = morphology_from_swc(infile)
    tree_list = morph.get_tree_list() 
    pair_idx = np.arange(0,len(tree_list))
    idx_list = []
    for i in pair_idx:
        select1 = [pair_data.index(p) for p in pair_data if p['tree1'] == i]
        select2 = [pair_data.index(p) for p in pair_data if p['tree2'] == i]      
        for s1 in select1:
            if pair_data[s1]['prob'] > thresh:
                if s1 not in idx_list:
                    cfactors = np.array([pair_data[s1]['cf0_near0'], pair_data[s1]['cf1_near0'],
                                        pair_data[s1]['cf0_far0'], pair_data[s1]['cf1_far0']])
                    if len(np.where(cfactors < 0.5)[0]):
                        print('idx %d'%(pair_data[s1]['idx']), cfactors)
                    else:
                        end_nodes = (morph.node_by_id(pair_data[s1]['nid1']),
                                    morph.node_by_id(pair_data[s1]['nid2']))
                        morph = connect_trees(end_nodes, morph)
                        idx_list.append(s1)
        for s2 in select2:
            if pair_data[s2]['prob'] > thresh:
                if s2 not in idx_list:
                    cfactors = np.array([pair_data[s2]['cf0_near0'], pair_data[s2]['cf1_near0'],
                                        pair_data[s2]['cf0_far0'], pair_data[s2]['cf1_far0']])
                    if len(np.where(cfactors < 0.5)[0]):
                        print('idx %d'%(pair_data[s2]['idx']), cfactors)
                    else:
                        end_nodes = (morph.node_by_id(pair_data[s2]['nid2']),
                                    morph.node_by_id(pair_data[s2]['nid1']))
                        morph = connect_trees(end_nodes, morph)
                        idx_list.append(s2)   
    morphology_to_swc(morph, outfile) 
    return morph 
            
class Reconnect(ags.ArgSchemaParser):
        
    def run(self):
        reconnect(self.args['input_file'], self.args['output_dir'], self.args['model_dir'],
        self.args['pxl_xyz'])
   
        
if __name__ == "__main__":
    mod = Reconnect(schema_type=ReconnectParameters)
    mod.run()      


import torch
from torch.autograd import Variable
import numpy as np
from ac_segmentation.neurotorch.datasets.dataset import Data


class Predictor:
    """
    A predictor segments an input volume into an output volume
    """
    def __init__(self, net, checkpoint, gpu_device=None):
        self.setNet(net, gpu_device=gpu_device)
        self.loadCheckpoint(checkpoint)

    def setNet(self, net, gpu_device=None):
        self.device = torch.device("cuda:{}".format(gpu_device)
                                   if gpu_device is not None
                                   else "cpu")

        self.net = net.to(self.device).eval()

    def getNet(self):
        return self.net

    def loadCheckpoint(self, checkpoint):
        self.getNet().load_state_dict(torch.load(checkpoint))

    def run(self, input_volume, output_volume, batch_size=6):
        self.setBatchSize(batch_size)

        with torch.no_grad():
            batch_list = [list(range(len(input_volume)))[i:i+self.getBatchSize()]
                          for i in range(0,
                                         len(input_volume),
                                         self.getBatchSize())]

            for batch_index in batch_list:
                batch = [input_volume[i] for i in batch_index]

                self.run_batch(batch, output_volume)

    def getBatchSize(self):
        return self.batch_size

    def setBatchSize(self, batch_size):
        self.batch_size = batch_size

    def run_batch(self, batch, output_volume):
        bounding_boxes, arrays = self.toTorch(batch)
        inputs = Variable(arrays).float()

        outputs = self.getNet()(inputs)
        
        tensor = torch.cat(outputs).data.cpu().numpy() # (B,3,128,128,128)
        tensor = np.uint8(np.round(255/(1 + np.exp(-tensor)))) # convert to probability and to uint8 before blend
        for ch in range(3):
            data_list = [Data(tensor[i][ch], bounding_box) for i, bounding_box in enumerate(bounding_boxes)]
            for data in data_list:
                output_volume[ch].blend(data)

    def toArray(self, data):
        torch_data = data.getArray().astype(np.float)
        torch_data = torch_data.reshape(1, 1, *torch_data.shape)
        return torch_data

    def toTorch(self, batch):
        bounding_boxes = [data.getBoundingBox() for data in batch]
        arrays = [self.toArray(data) for data in batch]
        arrays = torch.from_numpy(np.concatenate(arrays, axis=0))
        arrays = arrays.to(self.device)

        return bounding_boxes, arrays

    def toData(self, tensor_list, bounding_boxes):
        tensor = torch.cat(tensor_list).data.cpu().numpy()
        batch = [Data(tensor[i][0], bounding_box)
                 for i, bounding_box in enumerate(bounding_boxes)]

        return batch



from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
import random


class Brightness(Augmentation):
    def __init__(self, volume, max_relative_brightness=0.05, **kwargs):
        self.setRelativeBrightness(max_relative_brightness)
        super().__init__(volume, **kwargs)

    def augment(self, bounding_box):
        raw, label = self.getParent().get(bounding_box)
        augmented_raw, augmented_label = self.brightness_augmentation(raw,
                                                                      label)

        return (augmented_raw, augmented_label)

    def setFrequency(self, frequency):
        self.frequency = frequency

    def setRelativeBrightness(self, relative_brightness):
        self.relative_brightness = relative_brightness

    def brightness_augmentation(self, raw_data, label_data,
                                brightness_range=0.05):
        augmented_raw = raw_data.getArray()
        brightness = random.uniform(-brightness_range, brightness_range)
        augmented_raw = augmented_raw + brightness
        augmented_raw_data = Data(augmented_raw, raw_data.getBoundingBox())

        return augmented_raw_data, label_data



from numbers import Number


class Vector:
    """
    A basic vector data type
    """
    def __init__(self, *components: Number):
        """
        Initializes a vector
        :param components: Numbers specifying the components of the vector
        :type components: List of numbers
        """
        self.setComponents(components)

    def setComponents(self, components: list):
        """
        Set the components in a vector
        :param components: A list of numbers specifying the vector's 
components
        :type components: List of numbers
        """
        if not all(isinstance(x, Number) for x in components):
            raise ValueError("components must contain all numbers instead" +
                             " it contains {}".format(components))
        self.components = tuple(components)

    def getComponents(self) -> list:
        """
        Retrieves the components of a vector
        :return: A list of numbers specifying the vector's components
        :rtype: List of numbers
        """
        return self.components

    def getDimension(self) -> int:
        """
        Returns the dimension of the vector
        :return: The vector's dimension
        :rtype: int
        """
        return len(self.getComponents())

    def getNumpyDim(self) -> list:
        """
        Returns the size of the bounding box in row-major order (Z, Y, X)
        :return: Size of the bounding box in row-major order (Z, Y, X)
        :rtype: list
        """
        return self.getComponents()[::-1]

    def __getitem__(self, idx):
        if not isinstance(idx, int):
            raise IndexError("the index must be an positive integer")
        if idx < 0 or idx >= self.getDimension():
            raise IndexError("the index is out-of-bounds")

        return self.getComponents()[idx]

    def __add__(self, other):
        if not isinstance(other, Vector):
            raise ValueError("other must be a vector instead"
                             " it is {}".format(type(other)))

        if self.getDimension() != other.getDimension():
            raise ValueError("other must have the same dimension instead "
                             + "self is {} and other is {}".format(self,
                                                                   other))

        result = tuple(s + o for s, o in zip(self.getComponents(),
                                             other.getComponents()))

        return Vector(*result)

    def __mul__(self, other):
        if isinstance(other, Number):
            result = [s * other for s in self.getComponents()]

        elif isinstance(other, Vector):
            result = [s*o for s, o in zip(self.getComponents(),
                                          other.getComponents())]

        else:
            raise ValueError("other must be a number or a vector instead"
                             " it is {}".format(type(other)))

        return Vector(*result)

    def __div__(self, other):
        if isinstance(other, Number):
            return self*(1/other)

        if isinstance(other, Vector):
            result = Vector([1/o for o in other.getComponents()])
            result *= self

            return result

    def __sub__(self, other):
        return self+(other*-1)

    def __neg__(self):
        return self*-1

    def __eq__(self, other):
        if not isinstance(other, Vector):
            raise ValueError("other must be a vector instead"
                             " it is {}".format(type(other)))

        if self.getDimension() != other.getDimension():
            raise ValueError("other must have the same dimension")

        return all(r1 == r2 for r1, r2 in zip(self.getComponents(),
                                              other.getComponents()))

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "{}".format(self.getComponents())

    def __iter__(self):
        return iter(self.getComponents())


class BoundingBox:
    """
    A basic data type specifying a cube
    """
    def __init__(self, edge1: Vector, edge2: Vector):
        """
        Initializes a bounding box
        :param edge1: A vector specifying the first edge of the box
        :type edge1: Vector
        :param edge2: A vector specifying the second edge of the box
        :type edge2: Vector
        """
        self.setEdges(edge1, edge2)

    def setEdges(self, edge1: Vector, edge2: Vector):
        """
        Sets the edges of the bounding box
        :param edge1: A vector specifying the first edge of the box
        :type edge1: Vector
        :param edge2: A vector specifying the second edge of the box
        :type edge2: Vector
        """
        if not isinstance(edge1, Vector) or not isinstance(edge2, Vector):
            raise ValueError("edges must be vectors")

        if edge1.getDimension() != edge2.getDimension():
            raise ValueError("edges must have the same dimension instead"
                             + " edge 1 is {} and edge 2 is {}".format(edge1,
                                                                       edge2))

        self.edge1 = edge1
        self.edge2 = edge2

    def getEdges(self) -> tuple:
        """
        Returns the two edges for the bounding box
        :return: A tuple of two vectors containing the edges of the box
        :rtype: tuple
        """
        return (self.edge1, self.edge2)

    def getDimension(self) -> int:
        """
        Returns the dimension of the bounding box
        :return: The dimension of the bounding box
        :rtype: int
        """
        return self.getEdges()[0].getDimension()

    def getSize(self) -> Vector:
        """
        Returns the size of the bounding box as a vector
        :return: The size of the bounding box
        :rtype: Vector
        """
        edge1, edge2 = self.getEdges()
        chunk_size = edge2 - edge1

        return chunk_size

    def getNumpyDim(self) -> list:
        """
        Returns the size of the bounding box in row-major order (Z, Y, X)
        :return: Size of the bounding box in row-major order (Z, Y, X)
        :rtype: list
        """
        return self.getSize().getComponents()[::-1]

    def isDisjoint(self, other) -> bool:
        """
        Determines whether two bounding boxes are disjoint from each other
        :param other: The other bounding box for comparison
        :type other: BoundingBox
        :return: True if the bounding boxes are disjoint, false otherwise
        :rtype: bool
        """
        if not isinstance(other, BoundingBox):
            raise ValueError("other must be a vector instead other is "
                             "{}".format(type(other)))

        result = any(r1 > r2 for r1, r2 in zip(self.getEdges()[0],
                                               other.getEdges()[1]))
        result |= any(r1 < r2 for r1, r2 in zip(self.getEdges()[1],
                                                other.getEdges()[0]))

        return result

    def isSubset(self, other):
        """
        Determines whether the bounding box is a subset of the other
        :param other: The other bounding box for comparison
        :type other: BoundingBox
        :return: True if the bounding box is a subset of the other, false
 otherwise
        :rtype: bool
        """
        if not isinstance(other, BoundingBox):
            raise ValueError("other must be a vector instead other is "
                             "{}".format(type(other)))

        # Determines whether the first bounding box's components are
        # less than or equal to the other's components
        result = any(r1 <= r2 for r1, r2 in zip(self.getEdges()[1],
                                                other.getEdges()[1]))

        # Determines whether the first bounding box's components are
        # greater than or equal to the other's components
        result &= any(r1 >= r2 for r1, r2 in zip(self.getEdges()[0],
                                                 other.getEdges()[0]))

        return result

    def isSuperset(self, other):
        """
        Determines whether the bounding box is a super set of the other
        :param other: The other bounding box for comparison
        :type other: BoundingBox
        :return: True if the bounding box is a super set of the other, false
 otherwise
        :rtype: bool
        """
        return other.isSubset(self)

    def intersect(self, other):
        """
        Returns the bounding box given by the intersection of two
bounding boxes
        :param other: The other bounding box for the intersection
        :type other: BoundingBox
        :return: An bounding box intersecting the two bounding boxes
        :rtype: BoundingBox
        """
        if self.isDisjoint(other):
            raise ValueError("The bounding boxes must not be disjoint")

        # The first edge contains the largest components of the first
        # edge of the two bounding boxes
        edge1 = Vector(*map(lambda x, y: max(x, y),
                            self.getEdges()[0].getComponents(),
                            other.getEdges()[0].getComponents()))

        # The second edge contains the smallest components of the second
        # edge of the two bounding boxes
        edge2 = Vector(*map(lambda x, y: min(x, y),
                            self.getEdges()[1].getComponents(),
                            other.getEdges()[1].getComponents()))

        return BoundingBox(edge1, edge2)

    def __str__(self):
        edge1, edge2 = self.getEdges()
        return "Edge1: {}\nEdge2: {}".format(edge1, edge2)

    def __add__(self, other):
        if not isinstance(other, Vector):
            raise ValueError("other must be a vector instead other is "
                             + "{}".format(type(other)))

        edge1, edge2 = self.getEdges()
        result = BoundingBox(edge1 + other, edge2 + other)
        return result

    def __sub__(self, other):
        return self.__add__(other*-1)

    def __eq__(self, other):
        if not isinstance(other, BoundingBox):
            raise ValueError("other must be a BoundingBox")

        s_edge1, s_edge2 = self.getEdges()
        o_edge1, o_edge2 = other.getEdges()

        return (s_edge1 == o_edge1) and (s_edge2 == o_edge2)

    def __ne__(self, other):
        return not (self == other)



import os
import numpy as np
import pandas as pd
import argschema as ags
import json
from operator import add
import itertools
from collections import deque, defaultdict
from scipy import ndimage as ndi
from skimage.morphology import remove_small_objects, skeletonize_3d
import tifffile as tif
from ac_segmentation.deprecated.add_radius import add_radius_to_morph, n_nodes_up_and_down, n_nodes_up, n_nodes_down, bounding_box, add_missing_radius_vals

class PostprocessParameters(ags.ArgSchema):
    output_dir = ags.fields.OutputDir(required=True, description='Output directory')
    chunk_size = ags.fields.Int(dtype='int', required=False, default=416)
    overlap = ags.fields.Int(dtype='int', required=False, default=16)

def postprocess(outdir, chunk_size, overlap, threshold=0.2, size_threshold=2000):
    num_chunks = len([f for f in os.listdir(os.path.join(outdir, 'Segmentation')) 
                            if f.startswith('chunk')])
    
    savedir = os.path.join(outdir, 'swc_files')
    if not os.path.isdir(savedir):
        os.mkdir(savedir)
    df = pd.read_csv(os.path.join(outdir, 'inputs','bbox_deskewed.csv'))
    bb = df.bounding_box.values
    
    for n in range(num_chunks):
        stack = load_stack(os.path.join(outdir, 'Segmentation', 'chunk%02d'%n))
        
        # Crop stack to original size 
        new_stack_size = bb[5], bb[4], bb[3]  #zyx
        stack = stack[0:new_stack_size[0],0:new_stack_size[1],0:new_stack_size[2]]
        stack_size = stack.shape

        # Zero values below threshold
        stack[stack <= int(np.round(255*threshold))] = 0
        
        # Save nonzero pixels as csv file (x,y,z,I)
        z,y,x = np.nonzero(stack)
        I = stack[z,y,x]
        np.savetxt(os.path.join(outdir, 'Segmentation', 'chunk%02d.csv'%n), 
                   np.stack((x,y,z,I), axis=1), fmt='%u', delimiter=',', header='x,y,z,I')
        
        # Binarize stack based on threshold
        stack = (stack > int(np.round(255*threshold))).astype(np.uint8)

        # Label connected components
        s = ndi.generate_binary_structure(3,3)
        stack = ndi.label(stack,structure=s)[0].astype(np.uint16)
        num_cc = np.max(stack)

        if num_cc != 0:
            # Remove components smaller than size_threshold 
            stack = remove_small_objects(stack, min_size=size_threshold, connectivity=3)
            unique_labels, counts = np.unique(stack,return_counts=True)

            # Convert all connected component labels to 1
            stack = (stack > 0).astype(np.uint8)

            # Skeletonize stack
            stack = skeletonize_3d(stack)   
            
            # Label connected components
            s = ndi.generate_binary_structure(3,3)
            stack = ndi.label(stack,structure=s)[0].astype(np.uint16)
            num_cc = np.max(stack)
            
            # Create dict with xyz coordinates for each cc
            cc_dict = {}
            cc_range = range(1,num_cc+1)
            for cc in cc_range:
                cc_dict[cc] = {'X':[],'Y':[],'Z':[]}

            for j in range(stack_size[0]):
                img = stack[j,:,:]
                unique_labels = np.unique(img) #return indices and ignore 0 
                for l in unique_labels:
                    if l != 0:
                        idx = np.where(img==l)
                        [cc_dict[l]['Y'].append(coord) for coord in idx[0]]
                        [cc_dict[l]['X'].append(coord) for coord in idx[1]]
                        [cc_dict[l]['Z'].append(j) for coord in idx[0]]   

            chunkdir = os.path.join(savedir, 'chunk%02d'%n)
            if not os.path.isdir(chunkdir):
                os.mkdir(chunkdir)

            # Create swc file for each cc in cc_dict
            cc2swc(cc_dict, chunkdir) 

            # Add radius to swc file
            file_list = [f for f in os.listdir(chunkdir) if f.endswith('swc')] 
            file_list.sort()
            for f in file_list:
                try:
                    morph_in = morphology_from_swc(os.path.join(chunkdir, f))
                    dict_out = add_radius_to_morph(morph_in, np.stack((x,y,z,I), axis=1))
                    morphology_to_swc(dict_out['morph'], os.path.join(chunkdir, f))
                except:
                    print('error')
        
            
def load_stack(dirname):
    # Load image stack filenames
    filelist = [f for f in os.listdir(dirname) if f.endswith('.tif')] 
    filelist.sort()
    
    # Calculate stack size
    filename = os.path.join(dirname, filelist[0])
    img = tif.imread(filename)
    cell_stack_size = len(filelist), img.shape[0], img.shape[1]
        
    stack = np.zeros(cell_stack_size, dtype=img.dtype)
    for i, f in enumerate(filelist):
        filename = os.path.join(dirname, f)
        img = tif.imread(filename)
        stack[i,:,:] = img
        
    return stack         

def cc2swc(cc_dict, dirname):
    # Create swc for each cc in cc_dict 
    cc_range = range(1,len(cc_dict)+1)
    for cc in cc_range:
        try:
            coord_values = cc_dict[cc]
            component_coordinates = np.array([coord_values['X'],coord_values['Y'],coord_values['Z']]).T

            # Make a node dictionary for this con comp so we can lookup in the 26 node check step
            node_dict = {}
            count=0
            for c in component_coordinates:
                count+=1
                node_dict[tuple(c)] = count

            # 26 nodes to check in defining neighbors dict
            movement_vectors = [p for p in itertools.product([0,1,-1], repeat=3)]
            movement_vectors.remove((0,0,0))

            # Create neighbors dict and find root nodes (nodes with only one neighbor)
            root_node_dict = {}
            neighbors_dict = {}
            count = 0
            for node in component_coordinates:
                count+=1
                node_neighbors = [] 
                num_neighbors = 0
                for vect in movement_vectors:
                    node_to_check = tuple(list(map(add,tuple(node),vect)))
                    if node_to_check in node_dict.keys():
                        node_neighbors.append(node_to_check)
                if len(node_neighbors) == 1:
                    root_node_dict[tuple(node)] = count
                neighbors_dict[tuple(node)] = node_neighbors   

            # Set start node    
            start_node = min(root_node_dict, key=root_node_dict.get)
            start_nodes_parent = -1
            # Assign parent-child relation
            parent_dict = {}
            parent_dict[start_node] = start_nodes_parent

            queue = deque([start_node])
            while len(queue) > 0:
                current_node = queue.popleft()
                my_connections = neighbors_dict[current_node]
                for node in my_connections:
                    if node not in parent_dict:
                        parent_dict[node] = current_node
                        queue.append(node)
                    else:
                        p = 'Initial start node' if parent_dict[node] == start_nodes_parent else str([parent_dict[node]])

            # Number each node for swc
            ct=0
            big_node_dict = {}
            for j in parent_dict.keys():
                ct+=1
                big_node_dict[tuple(j)] = ct

            # Make swc list for swc file writing        
            node_type = 2 #axon
            swc_list = []
            for k,v in parent_dict.items():
                # id,type,x,y,z,r,pid
                if v == -1:
                    parent = -1
                else:
                    parent = big_node_dict[v]
                swc_line = [big_node_dict[k]] + [node_type] + list(k) + [1] + [parent]

                swc_list.append(swc_line)

            # Write swc file
            swc_file = os.path.join(dirname, '%05d.swc'%cc)
            with open(swc_file, 'w') as f:
                f.write('#id,type,x,y,z,r,pid\n')
                for i in range(len(swc_list)):
                    f.write('%d %d %d %d %d %.1f %d\n'%tuple(swc_list[i]))
        except:
            print('error')    

    
class Postprocess(ags.ArgSchemaParser):
        
    def run(self):
        postprocess(self.args['output_dir'], chunk_size=self.args['chunk_size'], 
        overlap = self.args['overlap'])   
        
if __name__ == "__main__":
    mod = Postprocess(schema_type=PostprocessParameters)
    mod.run()       


import torch
from torch.nn import BCEWithLogitsLoss, Module
import numpy as np
from scipy.ndimage.measurements import label


class SimplePointBCEWithLogitsLoss(Module):
    """
    Weights the binomial cross-entropy loss by the non-simple points
    """
    def __init__(self):
        super().__init__()
        self.bce = BCEWithLogitsLoss()

    def forward(self, prediction, label):
        weighted_prediction = self.simple_weight(prediction)
        weighted_label = self.simple_weight(label)

        cost = self.bce(weighted_prediction, weighted_label)

        return cost

    def simple_weight(self, tensor, simple_weight=1, non_simple_weight=10):
        non_simple_points = self.label_nonsimple_points(tensor)
        simple_points = tensor.new_ones(tensor.size()).to(tensor.get_device()) - \
                        non_simple_points
        inputs_weights = non_simple_weight * non_simple_points + \
                         simple_weight * simple_points
        result = inputs_weights * tensor
        return result

    def label_nonsimple_points(self, tensor, threshold=0.5):
        """
        Labels every non-simple point in a tensor

        :param tensor: A PyTorch tensor
        :param threshold: The threshold to binarize the tensor
        """
        try:
            device = tensor.get_device()
        except RuntimeError:
            raise RuntimeError("simple point weighting currently only works" +
                               " for GPUs")
        array = tensor.to("cpu")
        array = array.data.numpy()
        array = (array > threshold)
        labeled_array, num_features = label(array)
        size = labeled_array.shape
        padded_array = np.pad(labeled_array, (1,), 'edge')
        result = np.zeros(size)

        for k in range(0, size[0]):
            for j in range(0, size[1]):
                for i in range(0, size[2]):
                    if self._is_nonsimple_point(padded_array[k:k+3,
                                                             j:j+3,
                                                             i:i+3]):
                        result[k, j, i] = 1

        result = torch.from_numpy(result.astype(np.float32)).to(device)

        return result

    def _is_nonsimple_point(self, neighborhood):
        """
        Determines whether the center voxel in a labeled 3x3 neighborhood is simple

        :param neighborhood: A labeled 3x3 Numpy array
        """
        # Skip if the point is background
        if (neighborhood[1, 1, 1] == 0).any():
            return False

        # Setup neighborhood
        result = np.copy(neighborhood)
        center_point_label = result[1, 1, 1]

        # Create 18-neighborhood structure
        s = np.zeros((3, 3, 3))
        s[0, :, :] = np.array([[0, 1, 0],
                               [1, 1, 1],
                               [0, 1, 0]])
        s[1, :, :] = np.array([[1, 1, 1],
                               [1, 1, 1],
                               [1, 1, 1]])
        s[2, :, :] = np.array([[0, 1, 0],
                               [1, 1, 1],
                               [0, 1, 0]])

        # Calculates the topological number of the cavity
        result[result == 0] = -1
        labeled_array, num_features = label(result != center_point_label,
                                            structure=s)

        if num_features != 1:
            return True

        # Calculates the topological number of the component
        result = (result == center_point_label)
        result[1, 1, 1] = 0
        labeled_array, num_features = label(result,
                                            structure=np.ones((3, 3, 3)))

        if num_features != 1:
            return True

        # If the prior conditions are not satisfied, the point is simple
        return False



from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
from ac_segmentation.neurotorch.datasets.datatypes import Vector, BoundingBox
import random
import numpy as np


class Stitch(Augmentation):
    def __init__(self, volume, max_error=20, **kwargs):
        self.setMaxError(max_error)
        super().__init__(volume, **kwargs)

    def setMaxError(self, max_error):
        self.max_error = max_error

    def augment(self, bounding_box):
        # Get error and location
        error = random.randrange(2, self.max_error)
        x_len = bounding_box.getSize()[0]
        location = random.randrange(10, x_len - 10)

        # Get initial bounding box
        edge1, edge2 = bounding_box.getEdges()
        edge2 += Vector(20, error, 0)
        initial_bounding_box = BoundingBox(edge1, edge2)

        # Get data
        raw_data, label_data = self.getParent().get(initial_bounding_box)
        raw, label = (raw_data.getArray().copy(), label_data.getArray().copy())
        augmented_raw, augmented_label = self.stitch(raw, label,
                                                     location=location,
                                                     error=error)

        # Convert to the data format
        augmented_raw_data = Data(augmented_raw, bounding_box)
        augmented_label_data = Data(augmented_label, bounding_box)

        return (augmented_raw_data, augmented_label_data)

    def stitch(self, raw, label, location=20, error=3):
        # Initialize distorted raw volume and label
        z_len, y_len, x_len = raw.shape

        distorted_raw = np.zeros((z_len,
                                y_len-error,
                                x_len)).astype(raw.dtype)

        distorted_label = np.zeros((z_len,
                                    y_len-error,
                                    x_len)).astype(label.dtype)

        # Shear raw volume
        distorted_raw[:, :, location:] = raw[:, :-error, location:]
        distorted_raw[:, :, :location] = raw[:, error:, :location]

        distorted_label[:, :, location:] = label[:, :-error, location:]
        distorted_label[:, :, :location] = label[:, error:, :location]

        # Shear label
        fill_region = label[:, error//2:-error//2,
                            location-10:location+10]
        fill_region = self.shear3d(fill_region, shear=error, axis=1)

        distorted_label[:, :, location-10:location+10] = fill_region

        # Clip augmented raw volume and label
        distorted_raw = distorted_raw[:, :, :-20].astype(raw.dtype)
        distorted_label = distorted_label[:, :, :-20].astype(label.dtype)

        return distorted_raw, distorted_label

    def shear3d(self, volume, shear=20, axis=1):
        result = volume.copy()
        shift_list = np.around(np.linspace(-shear//2, shear//2, num=result.shape[2])).astype(volume.dtype)
        for index, shift in enumerate(shift_list):
            result[:, :, index] = np.roll(result[:, :, index], shift, axis=axis)

        return result



from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
from scipy.ndimage.filters import gaussian_filter
from numpy.random import normal
from scipy.ndimage.filters import median_filter
from random import random


class Blur(Augmentation):
    def __init__(self, volume, max_blur=(0.66, 4, 4), **kwargs):
        self.setMaxBlur(max_blur)
        super().__init__(volume, **kwargs)

    def augment(self, bounding_box):
        raw, label = self.getParent().get(bounding_box)
        augmented_raw, augmented_label = self.blur(raw, label, self.max_blur)

        return (augmented_raw, augmented_label)

    def setFrequency(self, frequency):
        self.frequency = frequency

    def setMaxBlur(self, max_blur):
        self.max_blur = max_blur

    def blur(self, raw_data, label_data, max_blur):
        raw = raw_data.getArray().copy()

        gaussian_raw = gaussian_filter(raw, sigma=max_blur)
        noise = raw - median_filter(raw, size=(3, 3, 3))

        gaussian_raw = gaussian_raw + noise
        gaussian_raw = gaussian_raw.astype(raw.dtype)

        augmented_raw_data = Data(gaussian_raw, raw_data.getBoundingBox())

        return augmented_raw_data, label_data



from ac_segmentation.neurotorch.core.trainer import TrainerDecorator
import torch
import os.path


class CheckpointWriter(TrainerDecorator):
    """
    A Trainer wrapper class to save training checkpoints
    """
    def __init__(self, trainer, checkpoint_dir, checkpoint_period=5000):
        """
        Initializes the checkpoint directory to write checkpoints

        :param trainer: Trainer object that the class wraps
        :param logger_dir: The directory to save checkpoints
        :param checkpoint_period: The number of iterations between checkpoints
        """
        if not os.path.isdir(checkpoint_dir):
            raise IOError("{} is not a valid directory".format(checkpoint_dir))

        super().__init__(trainer)

        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_period = checkpoint_period
        self.max_accuracy = 0

        self.iteration = 1
        print(self.getTrainer())

    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the checkpoint if there have been enough iterations

        :param sample_batch: A batch of input/label samples for training
        """
        loss = super().run_epoch(sample_batch)

        if self.iteration % self.checkpoint_period == 0:
            self.save_checkpoint("iteration_{}.ckpt".format(self.iteration))

        self.iteration += 1

        return loss

    def evaluate(self, batch):
        loss, accuracy, output = super().evaluate(batch)

        if accuracy > self.max_accuracy:
            self.max_accuracy = accuracy
            self.save_checkpoint("best.ckpt")

        return loss, accuracy, output

    def save_checkpoint(self, checkpoint_name):
        """
        Saves a training checkpoint
        """
        checkpoint_filename = os.path.join(self.checkpoint_dir,
                                           checkpoint_name)
        torch.save(self.getTrainer().net.state_dict(), checkpoint_filename)



# Kisuk Lee <kisuklee@mit.edu>, 2017
# https://github.com/seung-lab/DeepEM
import numpy as np
import torch
import torch.nn as nn
from ac_segmentation.neurotorch.utils import torch_utils

class EdgeSampler(object):
    def __init__(self, edges):
        self.edges = list(edges)

    def generate_edges(self):
        return list(self.edges)

    def generate_true_aff(self, obj, edge):
        o1, o2 = torch_utils.get_pair(obj, edge)
        ret = (((o1 == o2) + (o1 != 0) + (o2 != 0)) == 3)
        return ret.type(obj.type())

    def generate_mask_aff(self, mask, edge):
        m1, m2 = torch_utils.get_pair(mask, edge)
        return (m1 * m2).type(mask.type())


class EdgeCRF(nn.Module):
    def __init__(self, criterion, size_average=False, class_balancing=False):
        super(EdgeCRF, self).__init__()
        self.criterion = criterion
        self.size_average = size_average
        self.balancing = class_balancing

    def forward(self, preds, targets): 
        assert len(preds) == len(targets) 
        loss = 0
        for pred, target in zip(preds, targets):
            l = self.criterion(pred, target)
            loss += l

        return loss

    def class_balancing(self, target, mask):
        if not self.balancing:
            return mask
        dtype = mask.type()
        m_int = mask * torch.eq(target, 1).type(dtype)
        m_ext = mask * torch.eq(target, 0).type(dtype)
        n_int = m_int.sum().item()
        n_ext = m_ext.sum().item()
        if n_int > 0 and n_ext > 0:
            m_int *= n_ext/(n_int + n_ext)
            m_ext *= n_int/(n_int + n_ext)
        return (m_int + m_ext).type(dtype)


class AffinityLoss(nn.Module):
    def __init__(self, edges, criterion, size_average=False,
                 class_balancing=False):
        super(AffinityLoss, self).__init__()
        self.sampler = EdgeSampler(edges)
        self.decoder = AffinityLoss.Decoder(edges)
        self.criterion = EdgeCRF(
            criterion,
            size_average=size_average,
            class_balancing=class_balancing
        )

    def forward(self, preds, label):
        pred_affs = list()
        true_affs = list()
        edges = self.sampler.generate_edges()
        for i, edge in enumerate(edges):
            try:
                pred_affs.append(self.decoder(preds, i))
                true_affs.append(self.sampler.generate_true_aff(label, edge))
            except:
                raise
        return self.criterion(pred_affs, true_affs)

    class Decoder(nn.Module):
        def __init__(self, edges):
            super(AffinityLoss.Decoder, self).__init__()
            assert len(edges) > 0
            self.edges = list(edges)

        def forward(self, x, i):
            num_channels = x.size(-4)
            assert num_channels == len(self.edges)
            assert i < num_channels and i >= 0
            edge = self.edges[i]
            return torch_utils.get_pair_first(x[...,[i],:,:,:], edge)



#!/usr/bin/env python

import json
import pathlib

import argschema
import numpy

from ac_segmentation.postprocess.skeletonize_array import (
    label_binary_array, threshold_binarize_array,
    skeletonize_labeled_array_concurrent
)
from ac_segmentation.utils.io import (
    read_gzip_array, write_cv_skels_iter_tar
)


def run(probability_input_path, skeleton_output_path,
        probability_threshold=0.05, label_size_threshold=80,
        skeletonize_options=None):
    skeletonize_options = (
        {} if skeletonize_options is None
        else skeletonize_options
    )

    # explicitly uint8 to float32....
    prob_map = read_gzip_array(
        probability_input_path,
        preprocess_func=lambda x: (x.astype(numpy.float32) / 255.))

    binary_arr = threshold_binarize_array(
        prob_map, threshold=probability_threshold)
    labeled_arr, _ = label_binary_array(
        binary_arr, size_threshold=label_size_threshold)

    skels = skeletonize_labeled_array_concurrent(
        labeled_arr, **skeletonize_options
    )

    write_cv_skels_iter_tar(skeleton_output_path, skels)


class SkeletonizationOptions(argschema.schemas.DefaultSchema):
    n_jobs = argschema.fields.Int(required=False, allow_none=True)


class SkeletonizeProbabilitiesParameters(argschema.ArgSchema):
    probability_input = argschema.fields.InputFile(required=True)
    skeleton_output = argschema.fields.OutputFile(required=True)
    probability_threshold = argschema.fields.Float(
        required=False, default=0.05)
    label_size_threshold = argschema.fields.Int(required=False, default=80)
    skeletonize_options = argschema.fields.Nested(
        SkeletonizationOptions,
        required=False, default=None, allow_none=True)

    output_json = argschema.fields.OutputFile(required=False, allow_none=True)


class SkeletonizeProbabilitiesModule(argschema.ArgSchemaParser):
    default_schema = SkeletonizeProbabilitiesParameters

    def output(self, d):
        out_json = self.args.get("output_json")
        if out_json:
            pathlib.Path(out_json).parent.mkdir(parents=True, exist_ok=True)
            with open(out_json, "w") as f:
                json.dump(f, d)

    @property
    def skeletonize_options(self):
        return self.args["skeletonize_options"]    

    def run(self):
        run(
            self.args["probability_input"],
            self.args["skeleton_output"],
            probability_threshold=self.args["probability_threshold"],
            label_size_threshold=self.args["label_size_threshold"],
            skeletonize_options=self.skeletonize_options
        )
        self.output(self.args)


if __name__ == "__main__":
    mod = SkeletonizeProbabilitiesModule()
    mod.run()


__all__ = [
    "SkeletonizeProbabilitiesModule",
    "SkeletonizeProbabilitiesParameters"
]



from ac_segmentation.neurotorch.datasets.dataset import Volume, Array, Data
from ac_segmentation.neurotorch.datasets.datatypes import BoundingBox, Vector
from abc import abstractmethod
import fnmatch
import os.path
import h5py
import numpy as np
import tifffile as tif


class TiffVolume(Volume):
    def __init__(self, tiff_file, bounding_box: BoundingBox,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(128, 128, 32)),
                 stride: Vector=Vector(64, 64, 16)):
        """
        Loads a TIFF stack file or a directory of TIFF files and creates a
corresponding three-dimensional volume dataset
        :param tiff_file: Either a TIFF stack file or a directory
containing TIFF files
        :param chunk_size: Dimensions of the sample subvolume
        """
        # Set TIFF file and bounding box
        self.setFile(tiff_file)
        super().__init__(bounding_box, iteration_size, stride)

    def setFile(self, tiff_file):
        if os.path.isfile(tiff_file) or os.path.isdir(tiff_file):
            self.tiff_file = tiff_file
        else:
            raise IOError("{} was not found".format(tiff_file))

    def getFile(self):
        return self.tiff_file

    def get(self, bounding_box):
        return self.getArray().get(bounding_box)

    def __enter__(self):
        if os.path.isfile(self.getFile()):
            try:
                print("Opening {}".format(self.getFile()))
                array = tif.imread(self.getFile())

            except IOError:
                raise IOError("TIFF file {} could not be " +
                              "opened".format(self.getFile()))

        elif os.path.isdir(self.getFile()):
            tiff_list = os.listdir(self.getFile())
            tiff_list = filter(lambda f: fnmatch.fnmatch(f, '*.tif'),
                               tiff_list)

            if tiff_list:
                array = tif.TiffSequence(tiff_list).asarray()

        else:
            raise IOError("{} was not found".format(self.getFile()))

        array = Array(array, bounding_box=self.getBoundingBox(),
                      iteration_size=self.getIterationSize(),
                      stride=self.getStride())
        self.setArray(array)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.setArray(None)

    def _indexToBoundingBox(self, idx):
        return self.getArray()._indexToBoundingBox(idx)


class Hdf5Volume(Volume):
    def __init__(self, hdf5_file, dataset, bounding_box: BoundingBox,
                 iteration_size: BoundingBox=BoundingBox(Vector(0, 0, 0),
                                                         Vector(128, 128, 20)),
                 stride: Vector=Vector(64, 64, 10)):
        """
        Loads a HDF5 dataset and creates a corresponding three-dimensional
volume dataset

        :param hdf5_file: A HDF5 file path
        :param dataset: A HDF5 dataset name
        :param chunk_size: Dimensions of the sample subvolume
        """
        self.setFile(hdf5_file)
        self.setDataset(dataset)
        super().__init__(bounding_box, iteration_size, stride)

    def setFile(self, hdf5_file):
        self.hdf5_file = hdf5_file

    def getFile(self):
        return self.hdf5_file

    def setDataset(self, hdf5_dataset):
        self.hdf5_dataset = hdf5_dataset

    def getDataset(self):
        return self.hdf5_dataset

    def __enter__(self):
        if os.path.isfile(self.getFile()):
            with h5py.File(self.getFile(), 'r') as f:
                array = f[self.getDataset()].value
                array = Array(array, bounding_box=self.getBoundingBox(),
                              iteration_size=self.getIterationSize(),
                              stride=self.getStride())
                self.setArray(array)

    def __exit__(self, exc_type, exc_value, traceback):
        self.setArray(None)



import torch
import torch.optim as optim
import torch.nn as nn
from torch.autograd import Variable
import torch.cuda
import numpy as np
import os
import logging

class Trainer(object):
    """
    Trains a PyTorch neural network with a given input and label dataset using multiple cells simultaneously
    """
    def __init__(self, net, inputs_volume, labels_volume, checkpoint_dir, checkpoint_period=5000, 
                 logger_dir=None, checkpoint=None, optimizer=None, criterion=None, 
                 max_epochs=10, gpu_device=None, validation_split=0.2):
        """
        Sets up the parameters for training

        :param net: A PyTorch neural network
        :param inputs_volume: A list containing training and validation inputs
        :param labels_volume: A list containing training and validation corresponding labels
        :param checkpoint_dir: The directory to save checkpoints
        :param checkpoint_period: The number of iterations between checkpoints
        """
        self.max_epochs = max_epochs

        self.device = torch.device("cuda:{}".format(gpu_device)
                                   if gpu_device is not None
                                   else "cpu")

        self.net = net.to(self.device)

        if checkpoint is not None:
            self.net.load_state_dict(torch.load(checkpoint))

        if optimizer is None:
            self.optimizer = optim.Adam(self.net.parameters())
        else:
            self.optimizer = optimizer

        if criterion is None:
            self.criterion = nn.BCEWithLogitsLoss()
        else:
            self.criterion = criterion

        if gpu_device is not None:
            self.gpu_device = gpu_device
            self.useGpu = True

        self.inputs_volume = inputs_volume
        self.labels_volume = labels_volume 
        
        if not os.path.isdir(checkpoint_dir):
            raise IOError("{} is not a valid directory".format(checkpoint_dir))
        
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_period = checkpoint_period
        self.max_accuracy = 0
        
        if logger_dir is not None and not os.path.isdir(logger_dir):
            raise IOError("{} is not a valid directory".format(logger_dir))
            
        self.logger = logging.getLogger("Trainer")
        self.logger.setLevel(logging.INFO)    
        
        if logger_dir:
            file_handler = logging.FileHandler(os.path.join(logger_dir,
                                                            "training.log"))
            self.logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        self.logger.addHandler(console_handler) 
                
    def toTorch(self, arr):
        torch_arr = arr.astype(np.float)
        torch_arr = torch_arr.reshape(1, *torch_arr.shape)
        return torch_arr    
    
    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the checkpoint if there have been enough iterations

        :param sample_batch: A dictionary containing inputs and labels with the keys 
        "input" and "label", respectively
        """
                
        inputs = Variable(sample_batch[0]).float()
        labels = Variable(sample_batch[1]).float()

        inputs, labels = inputs.to(self.device), labels.to(self.device)

        self.optimizer.zero_grad()

        outputs = self.net(inputs)

        loss = self.criterion(torch.cat(outputs), labels)
        accuracy = torch.sum((torch.cat(outputs) > 0) & labels.byte()).float()
        accuracy /= torch.sum((torch.cat(outputs) > 0) | labels.byte()).float()
        loss_hist = loss.cpu().item()
        loss.backward()
        self.optimizer.step()
        
        return loss_hist, accuracy.cpu().item()
    
    def evaluate(self, batch):       
        
        with torch.no_grad():
            inputs = Variable(batch[0]).float()
            labels = Variable(batch[1]).float()

            inputs, labels = inputs.to(self.device), labels.to(self.device)

            outputs = self.net(inputs)

            loss = self.criterion(torch.cat(outputs), labels)
            accuracy = torch.sum((torch.cat(outputs) > 0) & labels.byte()).float()
            accuracy /= torch.sum((torch.cat(outputs) > 0) | labels.byte()).float()
            
            if accuracy > self.max_accuracy:
                self.max_accuracy = accuracy
                self.save_checkpoint("best.ckpt")
        
        return loss.cpu().item(), accuracy.cpu().item(), torch.stack(outputs).cpu().numpy()
    
    def run_training(self):
        """
        Trains the given neural network
        """
        num_epoch = 1
        num_iter = 1

        train_idx = np.random.permutation(len(self.inputs_volume[0]))
        train_idx = train_idx[:(len(train_idx) - len(train_idx) % 8)]
        train_idx = train_idx.reshape((-1, 8))
        val_idx = np.random.permutation(len(self.inputs_volume[1]))

        while num_epoch <= self.max_epochs:
            np.random.shuffle(train_idx)
            for i in range(train_idx.shape[0]):
                sample_batch = [np.stack([self.toTorch(self.inputs_volume[0][idx]) for idx in train_idx[i]]),
                                np.stack([self.toTorch(self.labels_volume[0][idx]) for idx in train_idx[i]])] 
                sample_batch[1] = sample_batch[1] > 0
                
                if num_epoch > self.max_epochs:
                    break

                print("Iteration: {}".format(num_iter))
                train_loss, train_acc = self.run_epoch([torch.from_numpy(batch.astype(np.float)) for batch in sample_batch])
                if num_iter % self.checkpoint_period == 0:
                    self.save_checkpoint("iteration_{}.ckpt".format(num_iter))
                
                if num_iter % 10 == 0:
                    val_batch = [np.stack([self.toTorch(self.inputs_volume[1][idx]) for idx in val_idx[:16]]),
                                 np.stack([self.toTorch(self.labels_volume[1][idx]) for idx in val_idx[:16]])] 
                    val_batch[1] = val_batch[1] > 0
                    loss, accuracy, _ = self.evaluate([torch.from_numpy(batch.astype(np.float)) for batch in val_batch])
                    self.logger.info("Iteration: {}, Epoch: {}/{}, Train loss: {:.4f}, Train acc: {:.2f}, Test loss: {:.4f}, Test acc: {:.2f}".format(num_iter, num_epoch, self.max_epochs, train_loss, train_acc*100, loss, accuracy*100))
                    
                num_iter += 1
        
            if num_epoch == self.max_epochs:
                while self.logger.handlers:
                    self.logger.handlers.pop()
            num_epoch += 1
    
    def save_checkpoint(self, checkpoint_name):
        """
        Saves a training checkpoint
        """
        checkpoint_filename = os.path.join(self.checkpoint_dir,
                                           checkpoint_name)
        torch.save(self.net.state_dict(), checkpoint_filename)


import importlib


class NetCollector(object):
    """
    Collects the different neural network architectures into a library
    """
    module_list = dict()

    def add_module(cls, module, identifier):
        NetCollector.module_list[identifier] = module

    def get_module(cls, identifier):
        importlib.import_module("neurotorch.nets." + identifier)

        try:
            return NetCollector.module_list[identifier]
        except KeyError:
            raise ValueError("{} could not be found in the" +
                             " NetCollector".format(identifier))



from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
from ac_segmentation.neurotorch.datasets.datatypes import Vector, BoundingBox
import random
import numpy as np
from scipy.ndimage import zoom


class Drop(Augmentation):
    def __init__(self, volume, max_slices=20, **kwargs):
        self.setMaxDroppedSlices(max_slices)
        super().__init__(volume, **kwargs)

    def augment(self, bounding_box):
        # Get dropped slices and location
        dropped_slices = 2*random.randrange(1, self.max_slices//2)
        location = random.randrange(dropped_slices,
                                    bounding_box.getSize()[1]-dropped_slices)

        # Get enlarged bounding box
        edge1, edge2 = bounding_box.getEdges()
        edge2 = edge2 + Vector(0, dropped_slices, 0)
        initial_bounding_box = BoundingBox(edge1, edge2)

        # Get data
        raw, label = self.getParent().get(initial_bounding_box)

        # Augment Numpy arrays
        augmented_raw, augmented_label = self.drop(raw, label,
                                                   dropped_slices=dropped_slices,
                                                   location=location)

        # Convert back into the data format
        augmented_raw_data = Data(augmented_raw, bounding_box)
        augmented_label_data = Data(augmented_label, bounding_box)

        return (augmented_raw_data, augmented_label_data)

    def setMaxDroppedSlices(self, max_slices):
        self.max_slices = max_slices

    def setLocation(self, location):
        self.location = location

    def drop(self, raw_data, label_data, dropped_slices=1, location=0):
        # Initialize distorted raw volume and label
        raw = raw_data.getArray().copy()
        label = label_data.getArray().copy()
        distorted_raw = np.zeros((raw.shape[0], raw.shape[1]-dropped_slices, raw.shape[2]))
        distorted_label = np.zeros((label.shape[0], label.shape[1]-dropped_slices, label.shape[2]))

        # Populate distorted raw volume and label
        distorted_raw[:, :location-dropped_slices//2, :] = raw[:, :location-dropped_slices//2, :]
        distorted_raw[:, location-dropped_slices//2:, :] = raw[:, location+dropped_slices//2:, :]

        distorted_label[:, :location-dropped_slices//2, :] = label[:, :location-dropped_slices//2, :]
        distorted_label[:, location-dropped_slices//2:, :] = label[:, location+dropped_slices//2:, :]

        # Interpolate the distorted label
        mag = 0.5
        fill_region = label[:, location-dropped_slices:location+dropped_slices, :]
        fill_region = zoom(fill_region, zoom=(1, mag, 1))
        fill_region = (fill_region > 0)

        # Fill in distorted label with interpolation
        distorted_label[:, location-dropped_slices:location, :] = fill_region

        return distorted_raw.astype(raw.dtype), distorted_label.astype(label.dtype)



import torch
import torch.optim as optim
import torch.nn as nn
from torch.autograd import Variable
from torch.utils.data import DataLoader
from ac_segmentation.neurotorch.loss.affinity import AffinityLoss
from ac_segmentation.neurotorch.utils import torch_utils
import torch.cuda
import numpy as np
import os
import logging

class Trainer(object):
    """
    Trains a PyTorch neural network with a given input and multilabel dataset using multiple cells simultaneously
    """
    def __init__(self, net, inputs_volume, labels_volume, augmentation, checkpoint_dir, checkpoint_period=1000, 
                 logger_dir=None, checkpoint=None, optimizer=None, criterion=None, 
                 max_epochs=10, gpu_device=None, validation_split=0.2):
        """
        Sets up the parameters for training

        :param net: A PyTorch neural network
        :param inputs_volume: A list containing training and validation inputs
        :param labels_volume: A list containing training and validation corresponding labels
        :param checkpoint_dir: The directory to save checkpoints
        :param checkpoint_period: The number of iterations between checkpoints
        """
        self.max_epochs = max_epochs

        self.device = torch.device("cuda:{}".format(gpu_device)
                                   if gpu_device is not None
                                   else "cpu")

        self.net = net.to(self.device)

        if checkpoint is not None:
            self.net.load_state_dict(torch.load(checkpoint, map_location=lambda storage, loc: storage.cuda(0))) # fix it loc: storage.cuda(gpu_device)

        if optimizer is None:
            self.optimizer = optim.Adam(self.net.parameters())
        else:
            self.optimizer = optimizer

        if criterion is None:
            edges = [(0,0,1),(0,1,0),(1,0,0)]
            criterion = nn.BCEWithLogitsLoss()
            self.criterion = AffinityLoss(edges,criterion) #nn.BCEWithLogitsLoss()
        else:
            self.criterion = criterion

        if gpu_device is not None:
            self.gpu_device = gpu_device
            self.useGpu = True

        self.inputs_volume = inputs_volume
        self.labels_volume = labels_volume 
        
        if not os.path.isdir(checkpoint_dir):
            raise IOError("{} is not a valid directory".format(checkpoint_dir))
        
        self.checkpoint_dir = checkpoint_dir
        self.checkpoint_period = checkpoint_period
        self.max_accuracy = 0
        
        if logger_dir is not None and not os.path.isdir(logger_dir):
            raise IOError("{} is not a valid directory".format(logger_dir))
            
        self.logger = logging.getLogger("Trainer")
        self.logger.setLevel(logging.INFO)    
        
        if logger_dir:
            file_handler = logging.FileHandler(os.path.join(logger_dir,
                                                            "training.log"))
            self.logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        self.logger.addHandler(console_handler) 
        
        self.augmentation = augmentation
                
    def toTorch(self, arr):
        torch_arr = arr.astype(np.float)
        torch_arr = torch_arr.reshape(1, *torch_arr.shape)
        return torch_arr    
    
    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the checkpoint if there have been enough iterations

        :param sample_batch: A list containing inputs and labels

        """
                
        inputs = Variable(sample_batch[0]).float()
        labels = Variable(sample_batch[1]).float()

        inputs, labels = inputs.to(self.device), labels.to(self.device)
        
        self.optimizer.zero_grad()

        outputs = self.net(inputs)
        
        loss = self.criterion(torch.cat(outputs), labels)
        accuracy = self.accuracy(torch.cat(outputs), labels)
        loss_hist = loss.cpu().item()
        loss.backward()
        self.optimizer.step()
        
        return loss_hist, accuracy.cpu().item()
    
    def evaluate(self, batch):       
        
        with torch.no_grad():
            inputs = Variable(batch[0]).float()
            labels = Variable(batch[1]).float()

            inputs, labels = inputs.to(self.device), labels.to(self.device)
            
            outputs = self.net(inputs)
            
            loss = self.criterion(torch.cat(outputs), labels)
            accuracy = self.accuracy(torch.cat(outputs), labels)
            
            if accuracy > self.max_accuracy:
                self.max_accuracy = accuracy
                self.save_checkpoint("best.ckpt")
        return loss.cpu().item(), accuracy.cpu().item(), torch.stack(outputs).cpu().numpy()
    
    def run_training(self):
        """
        Trains the given neural network
        """
        num_epoch = 1
        num_iter = 1
        batch_size = 3

        train_idx = np.random.permutation(len(self.inputs_volume[0]))
        train_idx = train_idx[:(len(train_idx) - len(train_idx) % batch_size)]
        train_idx = train_idx.reshape((-1, batch_size))
        val_idx = np.random.permutation(len(self.inputs_volume[1]))

        while num_epoch <= self.max_epochs:
            np.random.shuffle(train_idx)
            for i in range(train_idx.shape[0]):
                inputs_list = []
                labels_list = []
                for idx in train_idx[i]:
                    inputs = self.inputs_volume[0][idx]
                    labels = self.labels_volume[0][idx]
                    # apply augmentation (optional)
                    if self.augmentation:
                        inputs, labels = self.augment(inputs, labels)
                    inputs_list.append(self.toTorch(inputs)) 
                    labels_list.append(self.toTorch(labels))
                sample_batch = [np.stack(inputs_list), np.stack(labels_list)]
                                
                if num_epoch > self.max_epochs:
                    break

                print("Iteration: {}".format(num_iter))
                train_loss, train_acc = self.run_epoch([torch.from_numpy(batch.astype(np.float)) for batch in sample_batch])
                if num_iter % self.checkpoint_period == 0:
                    self.save_checkpoint("iteration_{}.ckpt".format(num_iter))
                
                if num_iter % 10 == 0:
                    np.random.shuffle(val_idx) # added shuffling to randomly sample validation set
                    val_batch = [np.stack([self.toTorch(self.inputs_volume[1][idx]) for idx in val_idx[:batch_size*2]]),
                                 np.stack([self.toTorch(self.labels_volume[1][idx]) for idx in val_idx[:batch_size*2]])] 
                    loss, accuracy, _ = self.evaluate([torch.from_numpy(batch.astype(np.float)) for batch in val_batch])
                    self.logger.info("Iteration: {}, Epoch: {}/{}, Train loss: {:.4f}, Train acc: {:.2f}, Test loss: {:.4f}, Test acc: {:.2f}".format(num_iter, num_epoch, self.max_epochs, train_loss, train_acc*100, loss, accuracy*100))
                    
                num_iter += 1
        
            if num_epoch == self.max_epochs:
                while self.logger.handlers:
                    self.logger.handlers.pop()
            num_epoch += 1
    
    def save_checkpoint(self, checkpoint_name):
        """
        Saves a training checkpoint
        """
        checkpoint_filename = os.path.join(self.checkpoint_dir,
                                           checkpoint_name)
        torch.save(self.net.state_dict(), checkpoint_filename)
        
    def accuracy(self, preds, label):
        edges = [(0,0,1),(0,1,0),(1,0,0)]
        pred_affs = list()
        true_affs = list()
        for i, edge in enumerate(edges):
            try:
                # generate pred aff
                num_channels = preds.size(-4)
                assert num_channels == len(edges)
                assert i < num_channels and i >= 0
                pred_affs.append(torch_utils.get_pair_first(preds[...,[i],:,:,:], edge))
                # generate true aff
                o1, o2 = torch_utils.get_pair(label, edge)
                ret = (((o1 == o2) + (o1 != 0) + (o2 != 0)) == 3)
                true_affs.append(ret.type(label.type()))      
            except:
                raise
        # calculate accuracy
        assert len(pred_affs) == len(true_affs)
        accuracy = 0
        for pred, target in zip(pred_affs, true_affs):
            acc = torch.sum((pred > 0) & (target > 0)).float()
            acc /= torch.sum((pred > 0) | (target > 0)).float()
            accuracy += acc
        accuracy /= len(pred_affs)
        return accuracy
        
    def augment(self, arr1, arr2, contrast_factor=0.3, brightness_factor=0.3):
        r = np.random.randint(4) # one of 4 orientations
        arr1 = np.rot90(arr1,r,axes=(1,2))
        arr2 = np.rot90(arr2,r,axes=(1,2))
        if np.random.randint(2): # y flip or not y flip
            arr1 = np.flip(arr1,1)
            arr2 = np.flip(arr2,1)
#         if np.random.randint(2): # z flip or not z flip
#             arr1 = np.flip(arr1,0)
#             arr2 = np.flip(arr2,0)
        # intensity perturbation 
        contrast = 1 + (np.random.rand() - 0.5) * contrast_factor # 1+[-0.15,0.15)
        brightness = (np.random.rand() - 0.5) * brightness_factor # [-0.15,0.15) 
        arr1 = arr1 * contrast
        arr1 = arr1 + brightness*65535
        arr1 = np.clip(arr1, 0, 65535)
        return arr1, arr2    
    


import torch
from torch.autograd import Variable
import numpy as np
import joblib
from joblib import Parallel, delayed
from ac_segmentation.neurotorch.datasets.dataset import Data
from ac_segmentation.utils.preprocess import lut_preprocess_array
from skimage import exposure

import ac_segmentation.neurotorch.datasets.dataset
from ac_segmentation.neurotorch.datasets.datatypes import BoundingBox, Vector
Array = ac_segmentation.neurotorch.datasets.dataset.Array
TSArray = ac_segmentation.neurotorch.datasets.dataset.TSArray



class Predictor:
    """
    A predictor segments an input volume into an output volume
    """
    def __init__(self, net, checkpoint, gpu_device=None):
        self.setNet(net, gpu_device=gpu_device)
        self.loadCheckpoint(checkpoint)

    def setNet(self, net, gpu_device=None):
        self.device = torch.device("cuda:{}".format(gpu_device)
                                   if gpu_device is not None
                                   else "cpu")

        self.net = net.to(self.device).eval()

    def getNet(self):
        return self.net

    def loadCheckpoint(self, checkpoint):
        self.getNet().load_state_dict(torch.load(checkpoint, map_location=self.device))

    def run(self, input_volume, output_volume, batch_size=30, mini_batch_size=10, max_pix = 65535):
        
        self.setBatchSize(batch_size)
        with torch.no_grad():
            batch_list = [list(range(len(input_volume)))[i:i+self.getBatchSize()]
                          for i in range(0,
                                         len(input_volume),
                                         self.getBatchSize())]

            for batch_index in batch_list:
                keep = []
                batch = [input_volume[i] for i in batch_index]
                if isinstance(input_volume, TSArray):
                    batch = [Data(np.pad(i.array.result(), pad_width=i.pad_size, mode="constant"),i.bounding_box) for i in batch]

                if hasattr(input_volume, 'mask'):
                    masks = [input_volume.mask[i] for i in batch_index]
                    if isinstance(input_volume.mask, TSArray):
                        masks = [Data(i.array.result(), i.bounding_box) for i in masks]
                    
                for ind,data in enumerate(batch):
                    if hasattr(input_volume, 'li_thresh'):
                        batch[ind].array[batch[ind].array < input_volume.li_thresh] = 0

                    if hasattr(input_volume, 'rescale_perc'):
                        p1, p2 = np.percentile(batch[ind].array, input_volume.rescale_perc)
                        batch[ind].array = exposure.rescale_intensity(batch[ind].array, in_range=(p1, p2))

                    if hasattr(input_volume, 'mask'):
                        batch[ind].array = np.where(masks[ind].array, batch[ind].array, 0)
                        
                    if np.any(data.array) == True:
                        if max_pix != None:
                            batch[ind].array = lut_preprocess_array(batch[ind].array, max_pix)
                        keep.append(batch[ind])

                if isinstance(input_volume, TSArray):
                    self.run_batch(keep, output_volume, input_volume.shift, mini_batch_size=mini_batch_size)

                else:
                    self.run_batch(keep, output_volume, mini_batch_size=mini_batch_size)

    def getBatchSize(self):
        return self.batch_size

    def setBatchSize(self, batch_size):
        self.batch_size = batch_size

    def run_batch(self, batch, output_volume, in_shift=[0,0,0], mini_batch_size=10):
        bounding_boxes, arrays = self.toTorch(batch)
        inputs = Variable(arrays).float()

        data_list = []
        n_list = range(len(inputs))
        batch_list = [n_list[i:i + mini_batch_size] for i in range(0, len(n_list), mini_batch_size)]  
        for s_batch in batch_list:
            st,end = int(s_batch[0]), int(s_batch[-1])
            outputs = self.getNet()(inputs[st:end])
            data_list += self.toData(outputs, bounding_boxes[st:end])

        if hasattr(output_volume, 'tensor'):
            writes = []
            for data in data_list:
                writes.append(output_volume.blend(data, in_shift))
                
            for write in writes:
                write.result()  

        else:
            for data in data_list:
                v1 = data.bounding_box.edge1 - Vector(*in_shift[::-1])
                v2 = data.bounding_box.edge2 - Vector(*in_shift[::-1])
                data = Data(data.array, BoundingBox(v1,v2))
                output_volume.blend(data)

    def toArray(self, data):
        torch_data = data.getArray().astype(float)
        torch_data = torch_data.reshape(1, 1, *torch_data.shape)
        return torch_data

    def toTorch(self, batch):
        bounding_boxes = [data.getBoundingBox() for data in batch]
        arrays = [self.toArray(data) for data in batch]
        arrays = torch.from_numpy(np.concatenate(arrays, axis=0))
        arrays = arrays.to(self.device)

        return bounding_boxes, arrays

    def toData(self, tensor_list, bounding_boxes):
        tensor = torch.cat(tensor_list).data.cpu().numpy()
        batch = [Data(tensor[i][0], bounding_box)
                 for i, bounding_box in enumerate(bounding_boxes)]

        return batch


import numpy as np
import tifffile as tif
import os
import glob
import argparse

import ac_segmentation.neurotorch.datasets.datatypes
import ac_segmentation.neurotorch.datasets.dataset
import ac_segmentation.neurotorch.nets.RSUNet
import ac_segmentation.neurotorch.core.predictor
import ac_segmentation.neurotorch.datasets.filetypes

Predictor = ac_segmentation.neurotorch.core.predictor.Predictor
Vector = ac_segmentation.neurotorch.datasets.datatypes.Vector
BoundingBox = ac_segmentation.neurotorch.datasets.datatypes.BoundingBox
Array = ac_segmentation.neurotorch.datasets.dataset.Array
RSUNet = ac_segmentation.neurotorch.nets.RSUNet.RSUNet
TiffVolume = ac_segmentation.neurotorch.datasets.filetypes.TiffVolume

def validate(checkpoint, test_dir, out_dir, bb, num_parts):
    # Initialize the U-Net architecture
    net = RSUNet()

    offset = 0
    for n in range(num_parts):
        bbn = BoundingBox(Vector(bb[0], bb[1], bb[2]), Vector(bb[3], bb[4], bb[5+n]))
        print(n, bbn)
        print('offset', offset) 
        if num_parts==1:
            filename = 'inputs' + '.tif'
        else:
            filename = 'inputs' + str(n) + '.tif'
        print(os.path.join(test_dir, filename)) 
        
        with TiffVolume(os.path.join(test_dir, filename), bbn) as inputs:              
            # Predict
            predictor = Predictor(net, checkpoint, gpu_device=0)
            output_volume = Array(-np.inf*np.ones(inputs.getBoundingBox().getNumpyDim(), dtype=np.float32))
            print('bb0', inputs.getBoundingBox())
            predictor.run(inputs, output_volume)
                            
            # Convert to probability map and save
            probability_map = 1/(1+np.exp(-output_volume.getArray()))
            print('probability_map', type(probability_map), probability_map.shape, probability_map.dtype)
            if not os.path.isdir(out_dir):
                os.mkdir(out_dir)
            for i in range(probability_map.shape[0]): # save 8-bit version as multiple tif files
                tif.imsave(os.path.join(out_dir,'%04d.tif'%(i+offset)), np.uint8(255*probability_map[i,:,:]))  
        offset = offset + bb[5+n]          

if __name__=="__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--ckpt', '-c', type=str, help='path to checkpoint')
    parser.add_argument('--test_dir', '-v', type=str, help='directory of validation/test data')
    parser.add_argument('--out_dir', '-o', type=str, help='results directory path')
    parser.add_argument('--bb', '-b', nargs='+', type=int, help='bounding box')
    parser.add_argument('--num_parts', '-n', type=int, help='number of parts to divide volume')
    args = parser.parse_args()
    validate(args.ckpt, args.test_dir, args.out_dir, args.bb, args.num_parts)



from ac_segmentation.neurotorch.augmentations.augmentation import Augmentation
from ac_segmentation.neurotorch.datasets.dataset import Data
import random
import numpy as np
from scipy.sparse import dok_matrix
from scipy.ndimage.filters import gaussian_filter


class Occlusion(Augmentation):
    def __init__(self, volume, **kwargs):
        super().__init__(volume, **kwargs)

    def augment(self, bounding_box):
        raw, label = self.getParent().get(bounding_box)
        augmented_raw, augmented_label = self.occlude(raw, label)

        return (augmented_raw, augmented_label)

    def occlude(self, raw_data, label_data, size=(4, 10, 10)):
        # Get array representations from data
        raw = raw_data.getArray()
        label = label_data.getArray()

        # Find a random neuron voxel
        neuron = self.dok_volume(label)
        neuron_voxel = random.choice(list(neuron))

        # Get occluding region
        min_x = max(0, neuron_voxel[2]-40)
        min_y = max(0, neuron_voxel[1]-40)
        min_z = max(0, neuron_voxel[0]-4)

        max_x = min(raw.shape[2], neuron_voxel[2]+40)
        max_y = min(raw.shape[1], neuron_voxel[1]+40)
        max_z = min(raw.shape[0], neuron_voxel[0]+4)

        x_len = max_x-min_x
        y_len = max_y-min_y
        z_len = max_z-min_z

        psf = np.zeros((z_len, y_len, x_len))
        psf[z_len//2, y_len//2, x_len//2] = 1
        psf = gaussian_filter(psf, size)
        psf = psf/psf[z_len//2, y_len//2, x_len//2]

        # Get background statistics
        sub_raw = raw[min_z:max_z, min_y:max_y, min_x:max_x]
        sub_label = label[min_z:max_z, min_y:max_y, min_x:max_x].astype(np.bool)
        average = np.mean(sub_raw[~sub_label])
        stdev = np.std(sub_raw[~sub_label])/2

        # Occlude region
        filtered_raw = raw.copy().astype(np.uint16)
        noise = np.random.normal(loc=average, scale=stdev, size=(z_len, y_len, x_len))
        filtered_raw[min_z:max_z, min_y:max_y, min_x:max_x] = (filtered_raw[min_z:max_z, min_y:max_y, min_x:max_x]*(1-psf) + \
                                                              psf*noise).astype(np.uint16)

        augmented_raw_data = Data(filtered_raw, raw_data.getBoundingBox())

        return augmented_raw_data, label_data

    def dok_volume(self, volume):
        dok = []
        for index in range(volume.shape[0]):
            dok += [(index, point[0], point[1]) for point in list(dok_matrix(volume[index, :, :]).keys())]
        return dok



#!/usr/bin/env python

# segment_and_skeletonize_strip.py updated with new features

import json
import pathlib

import argschema
import numpy

from ac_segmentation.postprocess.skeletonize_array import (
    label_binary_array, threshold_binarize_array,
    skeletonize_labeled_array_concurrent
)

from ac_segmentation.utils.io import (
    gzip_array, write_cv_skels_iter_tar
)
from ac_segmentation.process.segment_array import (
    predict_zarr_ts
)


import torch

# hack to set number of threads w/ OMP environment variable for old torch
import os
if os.getenv("OMP_NUM_THREADS"):
    torch.set_num_threads(int(os.getenv("OMP_NUM_THREADS")))

        
def run(weights_file, input_zarr, probability_output_path,
        skeleton_output_path, zarr_level=0, probability_threshold=0.05,
        label_size_threshold=80, filter_max_intensity=30000,
        predict_options=None, skeletonize_options=None):
        
    # predict and return as probability
    prob_map = predict_zarr_ts(
        input_zarr, weights_file, level=zarr_level,
        max_intensity=filter_max_intensity, gpu_device=predict_options['gpu_device'], 
        bound_box=predict_options['bound_box'])

    # write out uint8 representation of probabilities
    uint8_prob_map = (prob_map * 255).astype(numpy.uint8)
    gzip_array(probability_output_path, uint8_prob_map)
    del uint8_prob_map

    # binarize volume, label, and skeletonize
    binary_arr = threshold_binarize_array(
        prob_map, threshold=probability_threshold)
    labeled_arr, _ = label_binary_array(
        binary_arr, size_threshold=label_size_threshold)

    # skels = skeletonize_labeled_array(labeled_arr, **skeletonize_options)
    skels = skeletonize_labeled_array_concurrent(
        labeled_arr, n_jobs=skeletonize_options['n_jobs'])

    # write skeletons to swc zip
    # write_kimi_skels_tar(skeleton_output_path, skels)
    write_cv_skels_iter_tar(skeleton_output_path, skels)


class SegmentationPredictOptions(argschema.schemas.DefaultSchema):
    gpu_device = argschema.fields.Int(required=False, allow_none=True, default=None)
    batch_size = argschema.fields.Int(required=False, allow_none=True, default=80)
    bound_box = argschema.fields.List(argschema.fields.Int(),required=False, default='', allow_none=True)

class SkeletonizationOptions(argschema.schemas.DefaultSchema):
    n_jobs = argschema.fields.Int(required=False, allow_none=True)


class SegmentSkeletonizeZarrParameters(argschema.ArgSchema):
    input_zarr = argschema.fields.InputDir(required=True)
    weights_file = argschema.fields.InputFile(required=True)
    probability_output = argschema.fields.OutputFile(required=True)
    skeleton_output = argschema.fields.OutputFile(required=True)

    zarr_level = argschema.fields.Int(required=False, default=1)
    probability_threshold = argschema.fields.Float(
        required=False, default=0.05)
    filter_max_intensity = argschema.fields.Int(required=False, default=30000)
    label_size_threshold = argschema.fields.Int(required=False, default=80)
    predict_options = argschema.fields.Nested(
        SegmentationPredictOptions, required=False,
        default=None, allow_none=True)
    skeletonize_options = argschema.fields.Nested(
        SkeletonizationOptions, required=False, default=None, allow_none=True)

    output_json = argschema.fields.OutputFile(required=False, allow_none=True)


class SegmentSkeletonizeZarrModule(argschema.ArgSchemaParser):
    default_schema = SegmentSkeletonizeZarrParameters

    def output(self, d):
        out_json = self.args.get("output_json")
        if out_json:
            pathlib.Path(out_json).parent.mkdir(parents=True, exist_ok=True)
            with open(out_json, "w") as f:
                json.dump(f, d)
    
    @property
    def predict_options(self):
        if self.args["predict_options"]==None:
            return {'gpu_device':None,'bound_box':None}
        else:
            return self.args["predict_options"]

    @property
    def skeletonize_options(self):
        if self.args["skeletonize_options"]==None:
            return {'n_jobs':10}
        else:
            return self.args["skeletonize_options"] 

    def run(self):
        run(self.args["weights_file"], self.args["input_zarr"],
            self.args["probability_output"], self.args["skeleton_output"],
            self.args["zarr_level"],
            self.args["probability_threshold"],
            self.args["label_size_threshold"],
            self.args["filter_max_intensity"],
            self.predict_options,
            self.skeletonize_options)
        self.output(self.args)


if __name__ == "__main__":
    mod = SegmentSkeletonizeZarrModule()
    mod.run()

__all__ = [
    "SegmentSkeletonizeZarrModule",
    "SegmentSkeletonizeZarrParameters"]



#!/usr/bin/env
__doc__ = """

Basic building block layers for constructing nets implemented in PyTorch.

Nicholas Turner <nturner@cs.princeton.edu>, 2017
Based on a similar module by
Kisuk Lee <kisuklee@mit.edu>, 2016-2017
"""


import torch
import torch.nn as nn
from torch.nn import init
import math


def pad_size(ks, mode):

    assert mode in ["valid", "same", "full"]

    if mode == "valid":
        return (0, 0, 0)

    elif mode == "same":
        assert all([x % 2 for x in ks])
        return tuple(x // 2 for x in ks)

    elif mode == "full":
        return tuple(x - 1 for x in ks)


class Conv(nn.Module):
    """ Bare bones 3D convolution module w/ MSRA init """

    def __init__(self, D_in, D_out, ks, st, pd, bias=True):

        nn.Module.__init__(self)
        self.conv = nn.Conv3d(D_in, D_out, ks, st, pd, bias=bias)
        init.kaiming_normal_(self.conv.weight)
        if bias:
            init.constant_(self.conv.bias, 0)

    def forward(self, x):
        return self.conv(x)


class FactConv(nn.Module):
    """ Factorized 3D convolution using Conv"""

    def __init__(self, D_in, D_out, ks, st, pd, bias=True):

        nn.Module.__init__(self)
        if ks[0] > 1:
            self.factor = Conv(D_in, D_out, (1, ks[1], ks[2]),
                               (1, st[1], st[2]), (0, pd[1], pd[2]), bias=False)
            ks = (ks[0], 1, 1)
            st = (st[0], 1, 1)
            pd = (pd[0], 0, 0)

        else:
            self.factor = None

        self.conv = Conv(D_in, D_out, ks, st, pd, bias)

    def forward(self, x):

        if self.factor is not None:
            return self.conv(self.factor(x))
        else:
            return self.conv(x)


class ConvT(nn.Module):
    """ Bare Bones 3D ConvTranspose module w/ MSRA init """

    def __init__(self, D_in, D_out, ks, st, pd=(0, 0, 0), bias=True):

        nn.Module.__init__(self)
        self.conv = nn.ConvTranspose3d(D_in, D_out, ks, st, pd, bias=bias)
        init.kaiming_normal_(self.conv.weight)
        if bias:
            init.constant_(self.conv.bias, 0)

    def forward(self, x):
        return self.conv(x)


class ResizeConv(nn.Module):
    """ Upsampling followed by a Convolution """

    def __init__(self, D_in, D_out, ks, st, pd, bias=True, mode="nearest"):

        nn.Module.__init__(self)

        self.upsample = Upsample2D(scale_factor=2, mode=mode)
        self.conv = Conv(D_in, D_out, ks, st, pd, bias=bias)

    def forward(self, x):

        return self.conv(self.upsample(x))


class Upsample2D(nn.Module):

    def __init__(self, scale_factor, mode="nearest"):

        nn.Module.__init__(self)

        self.scale_factor = scale_factor
        self.upsample = nn.Upsample(scale_factor=2, mode=mode)

    def forward(self, x):

        # upsample in all dimensions, and undo the z upsampling
        return self.upsample(x)[:, :, ::self.scale_factor, :, :]


class FactConvT(nn.Module):
    """ Factorized 3d ConvTranspose using ConvT """

    def __init__(self, D_in, D_out, ks, st, pd=(0, 0, 0), bias=True):

        nn.Module.__init__(self)
        if ks[0] > 1:
            self.factor = ConvT(D_in, D_out, (2, ks[1], ks[2]),
                                (1, st[1], st[2]), (0, pd[1], pd[2]), bias=False)
            ks = (ks[0], 1, 1)
            st = (st[0], 1, 1)
            pd = (pd[0], 0, 0)

        else:
            self.factor = None

        self.conv = ConvT(D_in, D_out, ks, st, pd, bias)

    def forward(self, x):

        if self.factor is not None:
            return self.conv(self.factor(x))
        else:
            return self.conv(x)
